import pytest
import qolab.tableflow as tblfl
import pandas as pd


def test_table_load_noinputs():
    assert tblfl.loadInOutTables() == (None, None)
    assert tblfl.loadInOutTables(
        inputFileName=None, outputFileName="non_existing_file"
    ) == (None, None)


def test_wrong_comment_in_table_file_to_load():
    with pytest.raises(Exception):
        # should raise ParserError
        tblfl.loadInOutTables(
            inputFileName="tests/tableflow_test_data/tableIn1.csv",
            outputFileName=None,
            comment="%",
        )


def test_right_comment_in_table_file_to_load():
    tIn, tOut = tblfl.loadInOutTables(
        inputFileName="tests/tableflow_test_data/tableIn1.csv",
        outputFileName=None,
        comment="#",
    )
    assert isinstance(tIn, pd.core.frame.DataFrame)


def test_table_equality_with_no_output_file_name():
    tIn, tOut = tblfl.loadInOutTables(
        inputFileName="tests/tableflow_test_data/tableIn1.csv",
        outputFileName=None,
        comment="#",
    )
    assert isinstance(tIn, pd.core.frame.DataFrame)
    assert isinstance(tOut, pd.core.frame.DataFrame)
    assert tIn.equals(tOut)
    col0 = tIn.keys()[0]
    vBefore = tIn.at[0, col0]
    tIn.at[0, col0] = vBefore + 1
    assert not tIn.equals(tOut)


def test_table_load_with_in_out_file_names():
    # different filenames, same content for ease of testing
    tIn, tOut = tblfl.loadInOutTables(
        inputFileName="tests/tableflow_test_data/tableIn1.csv",
        outputFileName="tests/tableflow_test_data/tableOut1nonProcessed.csv",
        comment="#",
    )
    assert isinstance(tIn, pd.core.frame.DataFrame)
    assert isinstance(tOut, pd.core.frame.DataFrame)
    assert tIn.equals(tOut)

    # different filenames, different content
    tIn, tOut = tblfl.loadInOutTables(
        inputFileName="tests/tableflow_test_data/tableIn1.csv",
        outputFileName="tests/tableflow_test_data/tableOut1pariallyProcessed.csv",
        comment="#",
    )
    assert isinstance(tIn, pd.core.frame.DataFrame)
    assert isinstance(tOut, pd.core.frame.DataFrame)
    assert not tIn.equals(tOut)
    assert "out1" in tOut.columns
    assert "out1" not in tIn.columns


def test_for_existing_row():
    tbl1 = pd.DataFrame({"a": [1, 2, 3], "b": [1, 4, 6]})
    r = pd.Series({"a": 2, "b": 4})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 1


def test_for_existing_row_with_NA():
    # NA in both table and raw should return a hit
    tbl1 = pd.DataFrame({"a": [1, 2, 3], "b": [1, pd.NA, 6]})
    r = pd.Series({"a": 2, "b": pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 1

    # should insert new row
    tbl1 = pd.DataFrame({"a": [1, 2, 3], "b": [1, 4, 6]})
    r = pd.Series({"a": 2, "b": pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3

    # should insert new row
    tbl1 = pd.DataFrame({"a": [1, 2, 3], "b": [1, 4, 6]})
    r = pd.Series({"a": 2, "b": pd.NA})
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3


def test_for_nonexisting_row_and_its_insertion():
    tbl1 = pd.DataFrame({"a": [1, 2, 3], "b": [1, 4, 6]})
    r = pd.Series({"a": 2, "b": 10})
    assert len(tbl1) == 3
    assert tblfl.ilocRowOrAdd(tbl1, r) == 3
    assert len(tbl1) == 4


def test_isRedoNeeded():
    r = pd.Series({"a": 2, "b": 4, "c": pd.NA})
    assert not tblfl.isRedoNeeded(r, ["a", "b"])
    assert tblfl.isRedoNeeded(r, ["c"])
    assert tblfl.isRedoNeeded(r, ["non_existing"])
    assert not tblfl.isRedoNeeded(r, ["b", "c"])


def test_reflowTable():
    tIn, tOut = tblfl.loadInOutTables(
        inputFileName="tests/tableflow_test_data/tableIn1.csv",
        outputFileName="tests/tableflow_test_data/tableOut1pariallyProcessed.csv",
        comment="#",
    )
    # check for warnings
    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn, tOut)

    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn, tOut, postProcessedColums=["dummyName"])

    def frow(row):
        return row

    with pytest.warns(UserWarning):
        tblfl.reflowTable(tIn, tOut, process_row_func=frow)

    # now run reflow
    def frow(row, extraInfo=None):
        rowOut = row.copy()
        rowOut["out1"] = row["x"] * row["x"]
        return rowOut

    assert len(tIn) != len(tOut)
    tblfl.reflowTable(
        tIn, tOut, process_row_func=frow, postProcessedColums=["out1", "out2"]
    )
    assert len(tIn) == len(tOut)
    assert (tOut["out1"] == tOut["x"] * tOut["x"]).all()

    # check that reflow is done
    tOut.loc[tOut["x"] == 1, "out1"] = pd.NA
    tblfl.reflowTable(
        tIn, tOut, process_row_func=frow, postProcessedColums=["out1", "out2"]
    )
    assert (tOut["out1"] == tOut["x"] * tOut["x"]).all()

    # check that reflow is not reprocessed
    tOut.loc[tOut["x"] == 1, "out1"] = 12121  # crazy number
    tblfl.reflowTable(
        tIn, tOut, process_row_func=frow, postProcessedColums=["out1", "out2"]
    )
    assert (tOut.loc[tOut["x"] == 1, "out1"] == 12121).all()  # should not change

    # now we are forcing redo
    tOut.loc[tOut["x"] == 1, "out1"] = 12121  # crazy number
    tblfl.reflowTable(
        tIn,
        tOut,
        process_row_func=frow,
        postProcessedColums=["out1", "out2"],
        redo=True,
    )
    assert not (tOut.loc[tOut["x"] == 1, "out1"] == 12121).all()  # must not be the same
    assert (tOut["out1"] == tOut["x"] * tOut["x"]).all()



"""
Module: MKS Granville-Phillips 390 Micro-Ion Gauge Control
Description: This module provides a class to interface with the 
             MKS Granville-Phillips 390 Micro-Ion Gauge
             using pyvisa.

Author: Mykhailo Vorobiov
Email: mvorobiov@wm.edu
Date Created: 2024-10-22
Date Updated: 2024-11-05
"""

from qolab.hardware.scpi import SCPIinstr
import pyvisa
import numpy as np
import time


class MKS390(SCPIinstr):
    """
    A class for interfacing with the MKS Granville-Phillips 390 Micro-Ion Gauge
    for vacuum measurements between 1e-11 and 1e+3 Torr.
    """

    def __init__(self, resource, gauge_id="#02", *args, **kwds):
        """
        Initialize the vacuum gaguge class
        """
        super().__init__(resource, *args, **kwds)
        self.resource = resource

        # self.config["Device model"] = "MKS 390"
        self.resource.baud_rate = 19200
        self.resource.timeout = 5000
        self.resource.read_termination = "\r"
        self.resource.write_termination = "\r"

        self.write = self.resource.write
        self.read = self.resource.read
        self.read_raw = self.resource.read_raw

        self.ignition_status = False
        self.id = gauge_id

        # self._read_ignition_status()

    def enable_ignition(self, ig_status=False):
        self._read_ignition_status()
        if not (self.ignition_status == ig_status):
            try:
                # Sending a command
                command = self.id + "IG 0"
                self.write(command)
                self.ignition_status = ig_status
                print(
                    f"[INFO] Vacuum gauge ignition status changed to: {self.ignition_status}"
                )
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Failed to change vacuum gauge ignition status.\n\t{e}")
        else:
            print(
                f"[WARNING] No change to vacuum gauge ignition status: {self.ignition_status}"
            )

    def get_ignition_status(self):
        self._read_ignition_status()
        return self.ignition_status

    def _read_ignition_status(self):
        try:
            # Sending a command
            command = self.id + "IGS"
            self.write(command)
            time.sleep(0.01)
            self.ignition_status = bool(self.read_raw(8).strip().split()[1])
            return self.ignition_status
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Failed to read vacuum gauge ignition status.\n\t{e}")

    # @BasicInstrument.tsdb_append # The function does not work with the decorator (need to work on it later)
    def get_pressure(self):
        try:
            # Sending a command
            command = self.id + "RD"
            self.write(command)
            # Waiting and reading the response
            time.sleep(0.01)
            response = float(self.read_raw(14).strip().split()[1])
            return response
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Failed to get vacuum gauge reading.\n\t{e}")
            return np.nan()


'''    
    def close(self):
        """
        Close the connection to the vacuum gauge.
        """
        if self.gauge:
            try:
                self.gauge.close()
                self.logger.info("Connection to the vacuum gaguge closed.")
            except pyvisa.VisaIOError as e:
                self.logger.error(f"Error closing connection to the vacuum gauge: {e}")
'''
# Example usage
if __name__ == "__main__":
    import pyvisa

    rm = pyvisa.ResourceManager()

    gauge = MKS390(rm, "visa://192.168.194.15/ASRL13::INSTR", gauge_id="#02")
    pressure = gauge.get_reading()
    gauge.close()
    print(pressure)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import sys
import os

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "qolab"
copyright = "qolab developers"
author = "Eugeniy Mikhailov"
release = "0.32"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "autoapi.extension",
    "sphinx.ext.napoleon",
]
autoapi_dirs = ["../qolab"]
autoapi_keep_files = True

templates_path = ["_templates"]
exclude_patterns = []

sys.path.insert(0, os.path.abspath(".."))
autosummary_mock_imports = ["pandas", "universal_tsdb", "justpy", "ue9", "cachetools"]

napoleon_google_docstring = True
napoleon_numpy_docstring = True


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

smartquotes = False
# html_theme = 'alabaster'
# html_theme = 'sphinx_rtd_theme'
html_theme = "furo"
html_static_path = ["_static"]



from qolab.atoms.rubidium import Rubidium
import pytest
from pytest import approx


def test_pressure_calculation():
    # liquid Rb phase
    assert Rubidium().P(333) == approx(0.0015338076093146425)
    # solid Rb phase
    assert Rubidium().P(300) == approx(6.4998452973323e-05)


def test_density_calculation():
    rb = Rubidium(line=2)  # natural abundance by default
    T = 333  # above melting point
    assert rb.N(T, "87") == approx(9.284458474584362e16)
    assert rb.N(T, "85") == approx(2.407687273125237e17)
    assert rb.N(T, "all") == approx(3.3361331205836736e17)
    assert rb.N(T, "all") == (rb.N(T, "87") + rb.N(T, "85"))
    with pytest.raises(Exception):
        assert rb.N(T, "bogus")


def test_absorption_calculation():
    rb = Rubidium(line=2)  # natural abundance by default
    alpha_per_atom = rb.getTotalAphaPerAtomFn(T_typical=333)
    T = 333  # above melting point

    # affected by 85 and 87 Rb
    assert rb.TotalAlpha(-1.9e9, T) == approx(37.757349154367304)
    assert alpha_per_atom(-1.9e9) == approx(1.1321124e-16, abs=1e-21)
    # affected by 85 Rb only
    assert rb.TotalAlpha(1.67e9, T) == approx(311.8943644256937)
    assert alpha_per_atom(1.67e9) == approx(9.34654475e-16, abs=1e-21)
    # affected by 87 Rb only
    assert rb.TotalAlpha(4.14e9, T) == approx(102.59365294994637)
    assert alpha_per_atom(4.14e9) == approx(3.07492247e-16, abs=1e-21)



"""Basic multimeter classes to be included in hardware aware classes."""

from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument


class Multimeter(BasicInstrument):
    """
    Multimeter basic class.

    Intended to be part of the hardware aware class.
    """

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "Multimeter"
        self.config["Device model"] = (
            "Generic Multimeter generator Without Hardware interface"
        )
        self.config["FnamePrefix"] = "Multimeter"
        self.config["Device model"] = "Generic Multimeter Without Hardware interface"
        self.config["FnamePrefix"] = "multimeter"
        self.deviceProperties.update({})

    # Minimal set of methods to be implemented.
    def getVdc(self):
        """Report DC Voltage"""
        print("getVdc is not implemented")
        return None

    def getVac(self):
        """Report AC Voltage"""
        print("getVac is not implemented")
        return None

    def getAdc(self):
        """Report DC Current"""
        print("getAdc is not implemented")
        return None

    def getAac(self):
        """Report AC Current"""
        print("getAac is not implemented")
        return None

    def getResistance(self):
        """Report Resistance"""
        print("getResistance is not implemented")
        return None

    def getResistance4Wires(self):
        """Report Resistance with 4 wire method"""
        print("getResistance4Wires is not implemented")
        return None

    def getDiode(self):
        """Report Diode Voltage drop"""
        print("getDiode is not implemented")
        return None

    def getFreq(self):
        """Report Frequency"""
        print("getFreq is not implemented")
        return None


class MultimeterSCPI(SCPIinstr, Multimeter):
    """
    SCPI enabled basic multimeter.

    Intended to be part of the hardware aware class.

    Example
    -------
    >>> rm = pyvisa.ResourceManager()
    >>> MultimeterSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    pass

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        Multimeter.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



import os

from qolab.file_utils import infer_compression, save_table_with_header


def test_infer_compression():
    assert infer_compression("data.dat") is None
    assert infer_compression("data.unknown_ext") is None
    assert infer_compression("data.dat.gz") == "gzip"
    assert infer_compression("data.dat.bz") == "bzip"
    assert infer_compression("data.dat.bz2") == "bzip"


def test_save_table_with_header_match_filename_to_compression():
    import tempfile

    data = [[1, 3], [2, 5]]
    headerstr = ["header 1", "header 2"]
    with tempfile.TemporaryDirectory(prefix="qolab_test") as d:
        fname = os.path.join(d, "t.dat")
        assert save_table_with_header(
            fname,
            data,
            header=headerstr,
            compressionmethod=None,
            match_filename_to_compression=True,
        ) == (fname)
        assert save_table_with_header(
            fname,
            data,
            header=headerstr,
            compressionmethod="gzip",
            match_filename_to_compression=True,
        ) == (fname + ".gz")
        assert save_table_with_header(
            fname,
            data,
            header=headerstr,
            compressionmethod="bzip",
            match_filename_to_compression=True,
        ) == (fname + ".bz")

        # now cases when requested extension does not match the compression
        assert save_table_with_header(
            fname + ".gz",
            data,
            header=headerstr,
            compressionmethod=None,
            match_filename_to_compression=True,
        ) == (fname + ".gz")
        assert save_table_with_header(
            fname + ".bz",
            data,
            header=headerstr,
            compressionmethod="gzip",
            match_filename_to_compression=True,
        ) == (fname + ".bz.gz")
        assert save_table_with_header(
            fname + ".gz",
            data,
            header=headerstr,
            compressionmethod="bzip",
            match_filename_to_compression=True,
        ) == (fname + ".gz.bz")

        # do as I told cases, extension does not match compression
        assert save_table_with_header(
            fname + ".gz",
            data,
            header=headerstr,
            compressionmethod=None,
            match_filename_to_compression=False,
        ) == (fname + ".gz")
        assert save_table_with_header(
            fname,
            data,
            header=headerstr,
            compressionmethod="gzip",
            match_filename_to_compression=False,
        ) == (fname)
        assert save_table_with_header(
            fname,
            data,
            header=headerstr,
            compressionmethod="bzip",
            match_filename_to_compression=False,
        ) == (fname)



"""
Provide basic method to process data describing tables.

Created by Eugeniy E. Mikhailov 2024/05/27

The basic idea that we will have an *input* table
with data description and we (re)generate *output* table
based on the input table with processed rows.

If output table already have processed rows with entries different from NA
such rows are skipped.

Super handy for bulk processing data files where only a few parameters changed.
"""

import pandas as pd
import numpy as np
import warnings


def loadInOutTables(inputFileName=None, outputFileName=None, comment=None):
    """Load the input and the output tables from files.

    The output table loaded only if the corresponding file exists.
    Otherwise it is a clone of the input table.

    Parameters
    ==========
    inputFileName : path or string
        Path to the input table filename. If this file does not exists,
        return None for both tables.
    outputFileName : path or string or None
        Path to the output table filename. If such file does not exit,
        clone the input table to the output one.
    comment : string or None (default)
        String which indicates a comment in the input `csv` file.
        Usually it is either '#' or '%'. If set to None, internally changed to '#'.
    """
    if not inputFileName:
        return None, None

    if not comment:
        comment = "#"

    tIn = pd.read_csv(inputFileName, comment=comment)
    tIn.columns = tIn.columns.str.removeprefix(" ")
    # clean up leading white space in columns names

    try:
        tOut = pd.read_csv(outputFileName, comment=comment)
    except Exception:
        tOut = tIn.copy(deep=True)

    return tIn, tOut


def ilocRowOrAdd(tbl, row):
    """Find a row in a table (`tbl`) similar to a provided `row`.

    NA in both sets treated as a match.
    If similar 'row' not found in the table, insert it.

    `tbl` can have more columns than `row` but all columns of `row`
    must be present in `tbl`.

    i.e. we looking for the following match
    # (tSub == row) | (tSub.isna() & row.isna())
    """
    # pandas is extremely slow to work with selections/views
    # so we transform everything to numpy
    t = tbl.to_numpy()
    cols = tbl.columns.to_numpy()
    r = row.to_numpy()
    colsRow = row.keys()
    res = np.zeros(t.shape, dtype=bool)  # array filled with False

    # columns not in row are not part of comparison so we set them to True
    for idx, c in enumerate(cols):
        if c not in colsRow:
            res[:, idx] = True

    for idx, c in enumerate(colsRow):
        if c not in cols:
            warnings.warn(f"input table is missing column '{c}'")
            continue
        if pd.isna(r[idx]):
            res[:, cols == c] = pd.isna(t[:, cols == c])
            continue
        res[:, cols == c] = t[:, cols == c] == r[idx]
    match = res.all(axis=1)
    if np.any(match):
        return np.where(match)[0][0]
    # we need to create new row since tbl does not has it
    i = len(tbl)
    updateTblRowAt(tbl, i, row)
    return i


def updateTblRowAt(tbl, i, row):
    """Update row  with position 'i' in the table ('tbl') with values from 'row'."""
    for k in row.keys():
        tbl.at[i, k] = row[k]
    return


def isRedoNeeded(row, cols2check):
    """Check is Redo required in a given row.

    Redo is required if *all* required entries in 'cols2check' are NA
    or we are missing columns in cols2check list

    Parameters
    ==========
    row: pandas row
        row to perform check on
    cols2check: list of strings
        List of strings with column names which considered as generated outputs.
    """
    for c in cols2check:
        if c not in row.keys():
            return True
    if row[cols2check].isna().all():
        return True
    return False


def reflowTable(
    tIn,
    tOut,
    process_row_func=None,
    postProcessedColums=None,
    extraInfo=None,
    redo=False,
):
    """Reflow/update table tOut in place based on the inputs specified in table tIn.

    Effectively maps unprocessed rows to ``process_row_func``.

    Parameters
    ==========
    postProcessedColums : list of strings
        List of column names which need to be generated
    extraInfo : dictionary (optional)
        Dictionary of additional parameter supplied to ``process_row_func``
    process_row_func : function
        Function which will take a row from the input table and generate
        row with post processed entries (columns).
        Expected to have signature like:
        ``rowOut = process_row_func(rowIn, extraInfo=userInfo)``
    redo : True or False (default)
        Flag indicating if reflow is needed unconditionally
        (i.e. True forces reflow of all entries).
    """
    if not process_row_func:
        warnings.warn("process_row_func is not provided, exiting reflowTable")
        return
    if not postProcessedColums:
        warnings.warn("postProcessedColums are not provided, exiting reflowTable")
        return

    for index, rowIn in tIn.iterrows():
        iOut = ilocRowOrAdd(tOut, rowIn)
        rowOutBefore = tOut.iloc[iOut]

        if not (redo or isRedoNeeded(rowOutBefore, postProcessedColums)):
            continue

        # processing data describing row
        rowOut = process_row_func(rowOutBefore, extraInfo=extraInfo)
        updateTblRowAt(tOut, iOut, rowOut)



import logging
import pyvisa
import numpy as np
import qolab.tsdb as tsdb
import time
from qolab.hardware import BasicInstrument
from qolab.hardware.scope import SDS1104X
from qolab.hardware.rf_generator import AgilentE8257D
from qolab.hardware.lockin import SRS_SR865A
from qolab.hardware.i_server import I800

# this should be done before justpy is called or log formatter does not work
logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)

tmplogger = logging.getLogger("qolab.gui.web")
tmplogger.setLevel(logging.INFO)
tmplogger = logging.getLogger("qolab.tsdb")
tmplogger.setLevel(logging.INFO)
logger = logging.getLogger("Magnetometer")
logger.setLevel(logging.INFO)


def getConfig(apparatus):
    config = apparatus.config.copy()
    ai = apparatus.instruments
    for n, i in ai.items():
        config[n] = i.getConfig()
    return config


class Apparatus(BasicInstrument):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.setLockinSlope(0)

    @BasicInstrument.tsdb_append
    def getBCurrent(self):
        scope = self.instruments["scope.monitor"]
        ch = scope.getTrace(2, decimate=False)
        bCurrent = np.mean(ch.y.values)
        # mV -> mA
        return bCurrent

    def getErrAndEit(self):
        scope = self.instruments["scope.feedback"]
        # we need to freeze scope since this two values are taken on the same scope
        trg_mode = scope.getTriggerMode()
        scope.setTriggerMode("STOP")
        err = self.getErr()
        eit = self.getEIT()
        scope.setTriggerMode(trg_mode)
        return (err, eit)

    @BasicInstrument.tsdb_append
    def getErr(self):
        scope = self.instruments["scope.feedback"]
        ch1 = scope.getTrace(1, decimate=False)
        err = np.mean(ch1.y.values)
        return err

    @BasicInstrument.tsdb_append
    def getEIT(self):
        scope = self.instruments["scope.feedback"]
        ch3 = scope.getTrace(3, decimate=False)
        eit = np.mean(ch3.y.values)
        return eit

    @BasicInstrument.tsdb_append
    def setFeefback(self, val):
        freq0 = self.getRFreq()
        self.setRFreq(freq0 + val)

    @BasicInstrument.tsdb_append
    def getRFreq(self):
        rfgen = self.instruments["rfgen"]
        return rfgen.getFreqFixed()

    @BasicInstrument.tsdb_append
    def setRFreq(self, val):
        rfgen = self.instruments["rfgen"]
        rfgen.setFreqFixed(val)

    @BasicInstrument.tsdb_append
    def getRFAmplitude(self):
        rfgen = self.instruments["rfgen"]
        return rfgen.getRFAmplitude()

    @BasicInstrument.tsdb_append
    def setRFAmplitude(self, val):
        rfgen = self.instruments["rfgen"]
        return rfgen.setRFAmplitude(val)

    @BasicInstrument.tsdb_append
    def setLockinSlope(self, val):
        self.lockin_slope = val


if __name__ == "__main__":
    # TSDB logger setting
    tsdb_client = tsdb.Client(
        "influx", "http://lumus.physics.wm.edu:8428", database="qolab"
    )
    tsdb_ingester = tsdb.Ingester(
        tsdb_client, batch=11, measurement_prefix="VAMPIRE.HighPower"
    )

    # creating Apparatus with all instruments to be logged
    apparatus = Apparatus(
        tsdb_ingester=tsdb_ingester,
        device_nickname="magnetometer",
    )
    apparatus.config["Device type"] = "QOL VAMPIRE HighPower magnetometer"
    apparatus.config["Device model"] = "v0.1"
    apparatus.config["FnamePrefix"] = "magnetometer_eit"
    apparatus.config["SavePath"] = "/mnt/qol_grp_data/data.VAMPIRE.HighPower"

    app_nickname = apparatus.config["DeviceNickname"]

    logger.info("Accessing hardware")
    rm = pyvisa.ResourceManager()
    instr_scope = rm.open_resource("TCPIP::192.168.0.61::INSTR")
    scope_fdbk = SDS1104X(
        instr_scope,
        device_nickname=".".join([app_nickname, "scope.feedback"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_scope = rm.open_resource("TCPIP::192.168.0.62::INSTR")
    scope_mon = SDS1104X(
        instr_scope,
        device_nickname=".".join([app_nickname, "scope.monitor"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_rfgen = rm.open_resource("TCPIP::192.168.0.114::INSTR")
    rfgen = AgilentE8257D(
        instr_rfgen,
        device_nickname=".".join([app_nickname, "rfgen"]),
        tsdb_ingester=tsdb_ingester,
    )
    instr_lockin = rm.open_resource("TCPIP::192.168.0.51::INSTR")
    lockin = SRS_SR865A(
        instr_lockin,
        device_nickname=".".join([app_nickname, "lockin"]),
        tsdb_ingester=tsdb_ingester,
    )

    # adding instruments to apparatus
    apparatus.instruments = {}
    ai = apparatus.instruments
    ai["rfgen"] = rfgen
    ai["lockin"] = lockin
    ai["cellTemperatureController"] = I800(
        device_nickname=".".join([app_nickname, "cellTemperatureController"]),
        tsdb_ingester=tsdb_ingester,
    )
    # Do not add scope.feedback to apparatus.instruments it will create recursive loop
    # in the saved data file
    # ai['scope.feedback'] = scope_fdbk
    scope_fdbk.config["SavePath"] = "/mnt/qol_grp_data/data.VAMPIRE.HighPower"
    scope_fdbk.config["FnamePrefix"] = "scope_eit"
    # scope_fdbk.setRoll(False)
    # scope_fdbk.setRun(True)
    # scope_fdbk.setTimePerDiv(0.0005)
    # scope_fdbk.setChanVoltsPerDiv(1, 0.02)
    # scope_fdbk.setChanVoltsPerDiv(3, 0.02)
    ai["scope.monitor"] = scope_mon
    # scope_mon.setRoll(False)
    # scope_mon.setRun(True)
    # scope_mon.setTimePerDiv(0.0005)
    # scope_mon.setChanVoltsPerDiv(2, 0.002)
    # scope_mon.setChanVoltageOffset(2, -0.0726)

    scope_fdbk.config["tags"] = {}

    cfFreq = [6_833_980_000, 6_834_686_000, 6_835_393_000]
    m = [-2, 0, 2]
    # rfgen.setSweepCentralFreq(6_833_980_000)
    rfgen.setSweepCentralFreq(6_834_686_000)
    # rfgen.setSweepCentralFreq(6_835_393_000)

    for fr, m in zip(cfFreq, m):
        rfgen.setSweepCentralFreq(fr)
        print(f"time to settle for {m=} with {fr=}")
        time.sleep(10)
        print("settling done")

        # if we want to save new set of traces repeat this two commands
        scope_fdbk.config["tags"]["apparatus"] = getConfig(apparatus)
        scope_fdbk.save(maxRequiredPoints=1000)

    tsdb_ingester.commit()



from qolab.hardware.basic import BasicInstrument
from ._basic import MultimeterSCPI
from qolab.hardware.scpi import SCPI_PROPERTY
from pyvisa import constants as pyvisa_constants
import time


class HP_34401(MultimeterSCPI):
    r"""HP 34401 multimeter (same as Agilent 34401)

    Example
    -------

    >>> from qolab.hardware.multimeter.hp_34401 import HP_34401
    >>> rm = pyvisa.ResourceManager()
    >>> instr=rm.open_resource('ASRL/dev/ttyUSB0::INSTR')
    >>> multimeter = HP_34401(instr)
    >>> print("------ Header start -------------")
    >>> print(str.join("\n", multimeter.getHeader()))
    >>> print("------ Header ends  -------------")
    """

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "HP 34401"
        self.resource.read_termination = "\r\n"
        self.resource.baud_rate = 9600
        self.resource.data_bits = 8
        self.resource.parity = pyvisa_constants.Parity.none
        self.resource.stop_bits = pyvisa_constants.StopBits.one
        self.resource.timeout = 5000
        self.switchTime = 0.5  # switch time in seconds for Function/Measurement change
        self.deviceProperties.update({"Function"})

    rawFunction = SCPI_PROPERTY(
        scpi_prfx="SENSe:FUNCtion",
        ptype=str,
        doc="""
        Current measurement function (Voltmeter DC/AC, Currentmeter DC/AC, etc.
        Important! When assigning, string should be quoted, i.e. use:
            Function=\'"VOLT:DC"\'

        Possible values:
            "VOLTage:DC", same as "VOLT"
            "VOLTage:DC:RATio"
            "VOLTage:AC"
            "CURRent:DC", same as "CURR"
            "CURRent:AC"
            "RESistance" (2-wire ohms)
            "FRESistance" (4-wire ohms)
            "FREQuency"
            "PERiod"
            "CONTinuity"
            "DIODe"
        """,
    )

    def getFunction(self):
        rfunc = self.rawFunction
        if rfunc == '"VOLT"':
            return "Vdc"
        elif rfunc == '"VOLT:AC"':
            return "Vac"
        elif rfunc == '"CURR"':
            return "Adc"
        elif rfunc == '"CURR:AC"':
            return "Aac"
        elif rfunc == '"RES"':
            return "Resistance"
        elif rfunc == '"FRES"':
            return "Resistance4Wires"
        elif rfunc == '"DIOD"':
            return "Diode"
        elif rfunc == '"FREQ"':
            return "Frequency"
        elif rfunc == '"PER"':
            return "Period"
        elif rfunc == '"CONT"':
            return "Continuity"
        else:
            return "Unknown"

    def toRemote(self):
        self.write("SYSTem:REMote")

    def toLocal(self):
        self.write("SYSTem:LOCal")

    def isSensing(self, test):
        return test == self.rawFunction

    Reading = SCPI_PROPERTY(
        scpi_prfx="SYSTem:REMote; :READ",
        ptype=float,
        doc="Report current measurement",
        no_setter=True,
    )

    # HP 34401 is tricky:
    # when reading over serial interface the value of measurement/function,
    # the device need to be switched to REMOTE mode.
    # But this leaves screen blank and unusable.
    # So we have to do the dance with toRemote and toLocal.
    def getReadingWithSettingFunction(
        self, desired_function_string, function_internal_name
    ):
        """
        Get the required reading.

        But first check if the multimeter is set to do specific function/measurement.
        If the instrument set to do different function, set to the desired one.
        Note: the function name should be in quotes, i.e. "Volt".
        """
        if not self.isSensing(function_internal_name):
            self.rawFunction = desired_function_string
            time.sleep(self.switchTime)
        # self.toRemote() # this is done inside of Reading
        ret = self.Reading
        self.toLocal()
        return ret

    @BasicInstrument.tsdb_append
    def getVdc(self):
        return self.getReadingWithSettingFunction('"VOLTage:DC"', '"VOLT"')

    @BasicInstrument.tsdb_append
    def getVac(self):
        return self.getReadingWithSettingFunction('"VOLTage:AC"', '"VOLT:AC"')

    @BasicInstrument.tsdb_append
    def getAdc(self):
        return self.getReadingWithSettingFunction('"CURRent:DC"', '"CURR"')

    @BasicInstrument.tsdb_append
    def getAac(self):
        return self.getReadingWithSettingFunction('"CURRent:AC"', '"CURR:AC"')

    @BasicInstrument.tsdb_append
    def getResistance(self):
        return self.getReadingWithSettingFunction('"RESistance"', '"RES"')

    @BasicInstrument.tsdb_append
    def getResistance4Wires(self):
        return self.getReadingWithSettingFunction('"FRESistance"', '"FRES"')

    @BasicInstrument.tsdb_append
    def getDiode(self):
        return self.getReadingWithSettingFunction('"DIODe"', '"DIOD"')

    @BasicInstrument.tsdb_append
    def getFreq(self):
        return self.getReadingWithSettingFunction('"FREQuency"', '"FREQ"')

    @BasicInstrument.tsdb_append
    def getPeriod(self):
        return self.getReadingWithSettingFunction('"PERiod"', '"PER"')


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("ASRL/dev/ttyUSB0::INSTR")
    multimeter = HP_34401(instr)
    print("------ Header start -------------")
    print(str.join("\n", multimeter.getHeader()))
    print("------ Header ends  -------------")



"""
Module: GW INSTEK GPP-250-4.5 power supply unit control
Description: This module provides a class to interface with the GW INSTEK GPP250-4.5 power
             supply unit (PSU) using pyvisa for setting current and voltage on the PSU

Author: Mykhailo Vorobiov
Email: mvorobiov@wm.edu
Date: 2024-10-02
Update: 2024-11-06
"""

from qolab.hardware.power_supply._basic import PowerSupplySCPI
from qolab.hardware.basic import BasicInstrument

import pyvisa
import numpy as np


class PSW25045(PowerSupplySCPI):
    """
    A class for interfacing with GW INSTEK GPP250-4.5
    signle channel power supply unit.
    """

    def __init__(self, resource, *args, **kwds):
        """
        Initialize the PSU class.
        """
        super().__init__(resource, *args, **kwds)
        self.resource = resource

        self.config["Device model"] = "GWInstek PSW 250-4.5"
        self.numberOfChannels = 2
        self.deviceProperties.update({"OpMode"})
        self.resource.read_termination = "\n"

        self.MAX_CURRENT = 4.5  # Amp
        self.MAX_VOLTAGE = 250.0  # Volt
        self.enable_output(False)

    @property
    def idn(self):
        return self.get_idn()

    def get_idn(self):
        """
        Query the identification of the instrument.

        :return: Identifier string if query is successful or None otherwise.
        """
        try:
            return self.query("*IDN?").strip()
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error retrieving identification: {e}")
            return None

    @BasicInstrument.tsdb_append
    def get_out_current(self):
        """
        Query the current reading of the instrument.

        :return: Current reading in amps as float number and None if error occurs.
        """

        try:
            return float(self.query("MEAS:CURR?"))
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error querying the current reading: {e}")
            return np.nan()

    @BasicInstrument.tsdb_append
    def get_out_voltage(self):
        """
        Query the voltage reading of the instrument.

        :return: Voltage reading in volts as float number and None if error occurs.
        """

        try:
            return float(self.query("MEAS:VOLT?"))
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error querying the voltage reading: {e}")
            return np.nan()

    def get_out_power(self):
        """
        Query the power reading of the instrument.

        :return: Power reading in watts as float number and None if error occurs.
        """

        try:
            return float(self.query("MEAS:POWE?"))
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error querying the power reading: {e}")
            return np.nan()

    def set_current(self, current):
        """
        Set the target current output of the instrument.
        """

        if current <= self.MAX_CURRENT and current >= 0.0:
            try:
                self.write(f"SOUR:CURR {current}")
                print(f"Current is set to {current} V")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting the current target output: {e}")
        else:
            print(
                f"[WARNING] Target current must be between 0 and {self.MAX_CURRENT} Amps."
                f"\n\tThe traget current left unchanged."
            )

    def set_voltage(self, voltage):
        """
        Set the target voltage output of the instrument.
        """
        if voltage <= self.MAX_VOLTAGE and voltage >= 0.0:
            try:
                self.write(f"SOUR:VOLT {voltage}")
                print(f"[INFO] Voltage is set to {voltage} V")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting the voltage target output: {e}")
        else:
            print(
                f"[WARNING] Target voltage must be between 0 and {self.MAX_VOLTAGE} Volts."
                f"\n\tThe traget voltage left unchanged."
            )

    def enable_output(self, output_state=False):
        """
        Set the output state of the instrument ON or OFF.

        :param output_state: Boolean flag. Sets the output ON if 'True' or OFF if 'False'.
        """
        self.output_state = (
            output_state  # This is here to define out state (need some work on it)
        )
        try:
            if self.output_state:
                self.write("OUTP:STAT 1")
                self.output_state = output_state
                print("[INFO] Output is ON")
            else:
                self.write("OUTP:STAT 0")
                self.output_state = not (output_state)
                print("[INFO] Output is OFF")
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error toggle the output: {e}")



from qolab.hardware.basic import BasicInstrument
from ._basic import RFGeneratorSCPI


class AgilentE8257D(RFGeneratorSCPI):
    """Agilent E8257D RF generator.

    Note: Fixed frequency and Center frequency (of sweep)
    are different in this model.
    """

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.resource.read_termination = "\n"
        self.config["Device model"] = "Agilent E8257D"
        self.deviceProperties.update(
            {
                "RFPowerState",
                "RFAmplitude",
                "ModulationState",
                "FM1State",
                "FM1Source",
                "FM1ModulationDepth",
                "FM2State",
                "FM2Source",
                "FM2ModulationDepth",
                "FrequencyMode",  # sweep or continious
                "SweepCentralFreq",
                "SweepSpan",
            }
        )

    @BasicInstrument.tsdb_append
    def getFreqFixed(self):
        qstr = ":FREQuency:Fixed?"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setFreqFixed(self, freq):
        cstr = f":FREQuency:FIXED {freq}Hz"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getSweepCentralFreq(self):
        qstr = ":FREQuency:CENTer?"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setSweepCentralFreq(self, cfreq):
        cstr = f":FREQuency:CENTer {cfreq}Hz"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getSweepSpan(self):
        qstr = ":FREQuency:SPAN?"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setSweepSpan(self, span):
        cstr = f":FREQuency:SPAN {span}Hz"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def setSweep(self, cfreq=None, span=None):
        if cfreq is not None:
            self.setSweepCentralFreq(cfreq)
        if cfreq is not None:
            self.setSweepSpan(span)

    @BasicInstrument.tsdb_append
    def getModulationState(self):
        return int(self.query(":MODulation:STATe?"))

    @BasicInstrument.tsdb_append
    def setModulationState(self, val):
        self.write(f":MODulation:STATe {val}")

    @BasicInstrument.tsdb_append
    def getRFPowerState(self):
        return int(self.query(":OUTPut:STATe?"))

    @BasicInstrument.tsdb_append
    def setRFPowerState(self, val):
        self.write(f":OUTPut:STATe {val}")

    @BasicInstrument.tsdb_append
    def getRFAmplitude(self):
        return float(self.query(":POWer:AMPLitude?"))

    @BasicInstrument.tsdb_append
    def setRFAmplitude(self, val):
        self.write(f":POWer:AMPLitude {val}")

    @BasicInstrument.tsdb_append
    def getFM1ModulationDepth(self):
        return float(self.query(":FM1:Deviation?"))

    @BasicInstrument.tsdb_append
    def setFM1ModulationDepth(self, val):
        self.write(f":FM1:Deviation {val}")

    @BasicInstrument.tsdb_append
    def getFM2ModulationDepth(self):
        return float(self.query(":FM2:Deviation?"))

    @BasicInstrument.tsdb_append
    def setFM2ModulationDepth(self, val):
        self.write(f":FM2:Deviation {val}")

    @BasicInstrument.tsdb_append
    def getFM1Source(self):
        return str(self.query(":FM1:Source?"))

    @BasicInstrument.tsdb_append
    def setFM1Source(self, val):
        self.write(f":FM1:Source {val}")

    @BasicInstrument.tsdb_append
    def getFM2Source(self):
        return str(self.query(":FM2:Source?"))

    @BasicInstrument.tsdb_append
    def setFM2Source(self, val):
        self.write(f":FM2:Source {val}")

    @BasicInstrument.tsdb_append
    def getFM1State(self):
        return int(self.query(":FM1:State?"))

    @BasicInstrument.tsdb_append
    def setFM1State(self, val):
        self.write(f":FM1:State {val}")

    @BasicInstrument.tsdb_append
    def getFM2State(self):
        return int(self.query(":FM2:State?"))

    @BasicInstrument.tsdb_append
    def setFM2State(self, val):
        self.write(f":FM2:State {val}")

    @BasicInstrument.tsdb_append
    def getFrequencyMode(self):
        return str(self.query(":Frequency:Mode?"))

    @BasicInstrument.tsdb_append
    def setFrequencyMode(self, val):
        self.write(f":Frequency:Mode {val}")

    @BasicInstrument.tsdb_append
    def startFrequencySweep(self):
        self.setFrequencyMode("Sweep")

    def stopFrequencySweep(self):
        self.setFrequencyMode("FIXED")


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    instr = rm.open_resource("TCPIP::192.168.0.114::INSTR")
    rfgen = AgilentE8257D(instr)
    print("------ Header start -------------")
    print(str.join("\n", rfgen.getHeader()))
    print("------ Header ends  -------------")



import platform
import re
import os
from datetime import date


def filename2os_fname(fname):
    r"""Translate Windows or Linux filename to OS dependent style.
    Takes in account the notion of 'Z:' drive on different systems.

    In particular replaces Z: <==> /mnt/qol_grp_data  and \\ <==> /

    Example
    -------
    >>> filename2os_fname("Z:\\dir1\\dir2\\file")  # when run on Linux
    '/mnt/qol_grp_data/dir1/dir2/file'
    """
    if platform.system() == "Windows":
        fname = re.sub("/mnt/qol_grp_data", "Z:", fname)
    else:
        fname = re.sub("Z:", "/mnt/qol_grp_data", fname)
        fname = re.sub(r"\\", "/", fname)

    fname = os.path.normpath(fname)
    return fname


def get_runnum(data_dir):
    r"""Reads, increments data counter and saves it back in the provided `data_dir`.
    If necessary creates counter file and full path to it.

    Examples
    --------
    >>> get_runnum('Z:\\Ramsi_EIT\\data\\')
    >>> get_runnum('/mnt/qol_grp_data/data')
    """
    data_dir = filename2os_fname(data_dir)
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)
    if not os.path.isdir(data_dir):
        print(f"ERROR: cannot create directory for data: {data_dir}")
        print("Will use current dir for storage")
        data_dir = "."

    runnumpath = os.path.join(data_dir, "autofile")
    # convert to OS dependent way
    runnumpath = filename2os_fname(runnumpath)

    if not os.path.exists(runnumpath):
        os.mkdir(runnumpath)
    runnum_file = os.path.join(runnumpath, "runnum.dat")
    runnum_file = filename2os_fname(runnum_file)

    run_number = 0
    if os.path.isfile(runnum_file):
        with open(runnum_file, "r") as f:
            content = f.readlines()
            run_number = int(content[0])
            f.close()

    # Increment it and fold if needed
    run_number = run_number + 1
    # Important: we are using five digit counters to synchronize
    # with qol_get_next_data_file.m
    if run_number > 99999:
        run_number = 0

    with open(runnum_file, "w") as f:
        f.write(f"{run_number}")
        f.close()
    return run_number


def get_next_data_file(
    prefix,
    savepath,
    run_number=None,
    datestr=None,
    date_format="%Y%m%d",
    extension="dat",
):
    """Generate a filename according to a standard naming scheme
    fname = os.path.join(savepath, f'{prefix}_{datestr}_{run_number:05d}.{extension}')
    if run_number is missing, acquires it with `get_runnum( savepath )`
    """
    if run_number is None:
        run_number = get_runnum(savepath)
    today = date.today()
    if datestr is None:
        datestr = today.strftime(date_format)
    fname = os.path.join(savepath, f"{prefix}_{datestr}_{run_number:05d}.{extension}")
    return fname


def infer_compression(fname):
    """Infers compression algorithm from filename extension"""
    compression = None  # usual suspect
    b, fext = os.path.splitext(fname)
    if fext == ".gz":
        compression = "gzip"
    elif (fext == ".bz") or (fext == ".bz2"):
        compression = "bzip"
    return compression


def save_table_with_header(
    fname,
    data,
    header="",
    comment_symbol="%",
    skip_headers_if_file_exist=False,
    item_format="e",
    item_separator="\t",
    compressionmethod=None,
    compresslevel=9,
    match_filename_to_compression=True,
):
    r"""Saves output to CSV or TSV file with specially formatted header.

    The file is appended if needed.
    It is possible to compress output file.

    Parameters
    ----------
    fname : string, full path of the saved file.
    data : array type (python or numpy).
    header : list or array of header strings to put at the beginning of the record
    comment_symbol : prefix for the header lines, default is '%'.
        Note that headers are actually prefixed with <comment_symbol> and <space>.
        Historically it is chosen as '%' to make files compatible with Matlab `load`.
        '#' symbmol is also good choice.
    skip_headers_if_file_exist : True or False (default).
        When True skip addition of headers in already existing file.
        Useful when appending to file with the same headers.
    item_format : output format like in formatted strings, examples are 'e', '.15e', 'f'
    item_separator : how to separate columns, '\t' is default.
        Natural choices are either ',' (comma) or '\t' (tab).
    compressionmethod :  compression method
        - None : no compression (default)
        - gzip : gzip method of compression
        - bzip : bzip2 method of compression
    compresslevel : 0 to 9 (default)
        Compression level as it is defined for gzip in Lib/gzip.py
        - 0 : no compression at all
        - 9 : the highest compression (default)
    match_filename_to_compression: True (default) or False
        If True changes the filename suffix in accordance with compression method,
        e.g. 'data.dat' -> 'data.dat.gz' if compression is set to 'gzip',
        otherwise assumes that users know what they do.
        However, there is one important distinction to make compression enabling back
        compatible with old code which does not have place to set compression, but
        still want to use it.
        If compressionmethod is None, but inferred from a filename compression is not None,
        than we do what filename prescribes.
    """
    fname = filename2os_fname(fname)
    compression_infered = infer_compression(fname)
    if (compression_infered != compressionmethod) and match_filename_to_compression:
        if (compressionmethod is None) and (compression_infered is not None):
            compressionmethod = compression_infered
        elif compressionmethod == "gzip":
            fname += ".gz"
        elif compressionmethod == "bzip":
            fname += ".bz"
    file_exist_flag = os.path.exists(fname)
    item_format = str.join("", ["{", f":{item_format}", "}"])
    _open = open  # standard file handler
    if compressionmethod == "gzip":
        import gzip

        _open = lambda fname, mode: gzip.open(  # noqa: E731
            fname, mode=mode, compresslevel=compresslevel
        )
    if compressionmethod == "bzip":
        import bz2

        _open = lambda fname, mode: bz2.open(  # noqa: E731
            fname, mode=mode, compresslevel=compresslevel
        )
    with _open(fname, mode="ab") as f:
        if not (file_exist_flag and skip_headers_if_file_exist):
            for line in header:
                f.write(f"{comment_symbol} {line}\n".encode("utf-8"))
        if data is not None:
            for r in data:
                line = item_separator.join(map(item_format.format, r))
                f.write(line.encode("utf-8"))
                f.write("\n".encode("utf-8"))
        f.close()
    return fname



import time
from qolab.hardware.basic import BasicInstrument


class PID(BasicInstrument):
    """Proportional–integral–derivative controller.

    Calculates feedback based on gains, time, and previous error signal measurements.
    """

    def __init__(self, Gp=0, Gi=0, Gd=0, sign=1, enable=True, *args, **kwds):
        super().__init__(*args, **kwds)
        self.config["Device model"] = "Generic Software PID loop"
        self.config["Device type"] = "PID loop"
        self.config["FnamePrefix"] = "pid"
        self.deviceProperties.update({"Gp", "Gi", "Gd", "Sign", "Enable"})
        self.setGp(Gp)
        self.setGi(Gi)
        self.setGd(Gd)
        self.setSign(sign)
        self.setEnable(enable)

        self.reset()

    def __repr__(self):
        s = ""
        s += f"{self.__class__.__name__}("
        s += f"Gp={self.Gp}"
        s += f", Gi={self.Gi}"
        s += f", Gd={self.Gd}"
        s += f", sign={self.enable}"
        if self.config["DeviceNickname"] is not None:
            s += ", device_nickname='"
            s += f"{self.config['DeviceNickname']}"
            s += "'"
        s += ")"
        return s

    @BasicInstrument.tsdb_append
    def getGp(self):
        return self.Gp

    @BasicInstrument.tsdb_append
    def setGp(self, val):
        self.Gp = val

    @BasicInstrument.tsdb_append
    def getGi(self):
        return self.Gi

    @BasicInstrument.tsdb_append
    def setGi(self, val):
        self.Gi = val

    @BasicInstrument.tsdb_append
    def getGd(self):
        return self.Gd

    @BasicInstrument.tsdb_append
    def setGd(self, val):
        self.Gd = val

    @BasicInstrument.tsdb_append
    def getSign(self):
        return self.sign

    @BasicInstrument.tsdb_append
    def setSign(self, val):
        self.sign = val

    @BasicInstrument.tsdb_append
    def getEnable(self):
        return self.enable

    @BasicInstrument.tsdb_append
    def setEnable(self, val):
        self.enable = val

    def reset(self):
        self.err_1dt_back = 0
        self.err_2dt_back = 0
        self.err_now = 0
        self.last_update = time.time()

    def feedback(self, err):
        # PID feedback
        # see https://en.wikipedia.org/wiki/PID_controller#Pseudocode
        self.err_2dt_back = self.err_1dt_back
        self.err_1dt_back = self.err_now
        self.err_now = err

        tnow = time.time()
        dt = tnow - self.last_update
        self.last_update = tnow
        A0 = self.Gp + self.Gi * dt + self.Gd / dt
        A1 = -self.Gp - 2 * self.Gd / dt
        A2 = self.Gd / dt

        u = A0 * self.err_now + A1 * self.err_1dt_back + A2 * self.err_2dt_back
        u *= self.sign
        if not self.enable:
            return 0
        return u



import yaml
import time
from qolab.file_utils import get_next_data_file
from qolab.tsdb import tsdb_append_metric_for_class_setter_or_getter


class BasicInstrument:
    """This is the most basic instrument class.

    It is intended to be used as a parent class for real instruments.

    Notable feature that when ``getConfig`` is called,
    it gathers values of the properties listed in ``deviceProperties`` variable and
    also if ``numberOfChannels`` is set, it gathers configs of channels listed in
    ``channelProperties``. The later case is intended for instruments with
    global setting and channels, for example scopes. This is super handy
    to include together with acquired and saved data to see under which condition
    they are taken.

    Some property might have ``tsdb_append`` decorator
    which sends the property value to the time
    series database (TSDB) for the log keeping. Use it only for properties
    which are scalar, for example VoltsPerDiv, SampleRate, and similar.
    Such logs could be later analyzed with a help of systems like Grafana
    or anything which could read TSDB entries.

    Parameters
    ----------
    device_nickname : None (default) or str
        used to distinguish similar (in hardware, e.g. scopes) instruments,
        also used as a tag in Time Series DB (TSDB) logging.
        If not set we fallback to 'Device type' in config.
    tsdb_ingester : None or tsdb ingester
        used to log properties (with setter/getter) to TSDB.
        If 'None', the log entry is skipped
    config : dictionary, default is empty dictionary
        used to add or override default entries describing instrument
        It is good idea to set the following keys in the dictionary:
        'Device type', 'Device model', 'DeviceId', 'DeviceNickname',
        'FnamePrefix', 'SavePath'.

    Example
    -------
    >>> config['Device type'] = 'Basic Instrument'
    >>> config['Device model'] = 'Model is unset'
    >>> config['DeviceId'] = None
    >>> config['DeviceNickname'] = device_nickname; # to separate similar instruments
    >>> config['FnamePrefix'] = 'basicInstrument'
    >>> config['SavePath'] = './data'
    """

    def __init__(self, config={}, device_nickname=None, tsdb_ingester=None):
        self.config = {}
        self.config["Device type"] = "Basic Instrument"
        self.config["Device model"] = "Model is unset"
        self.config["DeviceId"] = None
        self.config["DeviceNickname"] = device_nickname
        # to separate similar instruments
        self.config["FnamePrefix"] = "basicInstrument"
        self.config["SavePath"] = "./data"
        for k, v in config.items():
            self.config[k] = v
        self.tsdb_ingester = tsdb_ingester
        # deviceProperties must have 'get' and preferably 'set' methods available,
        # i.e. 'SampleRate' needs getSampleRate() and love to have setSampleRate(value)
        # they will be used to obtain config and set device according to it
        # self.deviceProperties = {'SampleRate', 'TimePerDiv', 'TrigDelay', };
        self.deviceProperties = {"TimeStamp"}

    def __repr__(self):
        s = ""
        s += f"{self.__class__.__name__}("
        s += "config={"
        cstr = []
        for k, v in self.config.items():
            if v is not None:
                cstr.append(f"'{k}'" ": " f"'{v}'")
        s += ", ".join(cstr)
        s += "}"
        if self.tsdb_ingester is not None:
            s += f", tsdb_ingester={self.tsdb_ingester}"
        s += ")"
        return s

    def tsdb_append(f):
        """Append value to TSDB. Intended as decorator for setters and getters."""
        return tsdb_append_metric_for_class_setter_or_getter()(f)

    def getChannelsConfig(self):
        chconfig = {}
        if not hasattr(self, "numberOfChannels"):
            return chconfig

        for chNum in range(1, self.numberOfChannels + 1):
            chNconfig = {}
            for p in self.channelProperties:
                getter = f"getChan{p}"
                if not hasattr(self, getter):
                    print(f"warning no getter for {p}, i.e. {getter} is missing")
                    continue
                res = getattr(self, getter)(chNum)
                chNconfig[p] = res
            chconfig[chNum] = chNconfig
        return chconfig

    def getConfig(self):
        config = self.config.copy()
        dconfig = {}
        for p in self.deviceProperties:
            if hasattr(self, p) or hasattr(type(self), p):
                dconfig[p] = getattr(self, p)
                continue
            getter = f"get{p}"
            if not hasattr(self, getter):
                print(f"warning no getter for {p}, i.e. {getter} is missing")
                continue
            res = getattr(self, getter)()
            dconfig[p] = res
        config["DeviceConfig"] = dconfig
        if not hasattr(self, "numberOfChannels"):
            return config
        config["ChannelsConfig"] = self.getChannelsConfig()
        return config

    def setConfig(self, cfg):
        new_config = cfg.copy()
        device_config = new_config.pop("DeviceConfig")
        self.config.update(new_config)
        for p, v in device_config.items():
            setter = f"set{p}"
            if hasattr(self, setter):
                # we have something like setProperty
                getattr(self, setter)(v)
                continue
            if hasattr(self, p) or hasattr(type(self), p):
                # we have attribute Property
                setattr(self, p, v)
                continue
            print(f"warning: both {setter} and attribute {p} are missing, skipping {p}")
            # self.deviceProperties.add(p)
            # setattr(self, p, v)

    def getTimeStamp(self):
        return time.strftime("%Y/%m/%d %H:%M:%S")

    def getHeader(self):
        header = yaml.dump(self.getConfig(), default_flow_style=False, sort_keys=False)
        header = header.split("\n")
        return header

    def getNextDataFile(self, extension="dat"):
        fname = get_next_data_file(
            self.config["FnamePrefix"], self.config["SavePath"], extension=extension
        )
        return fname



"""
Created by Eugeniy E. Mikhailov 2024/07/18
"""

from qolab.hardware.basic import BasicInstrument
from qolab.hardware.scpi import SCPI_PROPERTY
from ._basic import ScopeSCPI, calcSparsingAndNumPoints
from qolab.data.trace import Trace
import numpy as np
import scipy.signal
from pyvisa.errors import VisaIOError
import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class RigolDS1054z(ScopeSCPI):
    """Rigol 1054 scope"""

    vertDivOnScreen = 8
    horizDivOnScreen = 12

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "Rigol DS1054z"
        self.deviceProperties.update({"MemoryDepth"})
        self.resource.read_termination = "\n"
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1200
        self.resource.timeout = 1000  # timeout in ms
        # desired number of points per channel, can return twice more

    TimePerDiv = SCPI_PROPERTY(
        scpi_prfx=":TIMEBASE:MAIN:SCALE",
        ptype=float,
        doc="Scope Time per Division",
    )

    def getTimePerDiv(self):
        return self.TimePerDiv

    def setTimePerDiv(self, value):
        self.TimePerDiv = value

    TrigDelay = SCPI_PROPERTY(
        scpi_prfx=":TIMEBASE:MAIN:OFFSET",
        ptype=float,
        doc="Scope Time Offset or Trigger Delay",
    )

    @BasicInstrument.tsdb_append
    def getTrigDelay(self):
        return self.TrigDelay

    @BasicInstrument.tsdb_append
    def setTrigDelay(self, value):
        self.TrigDelay = value

    @BasicInstrument.tsdb_append
    def getChanVoltageOffset(self, chNum):
        qstr = f":CHANnel{chNum}:OFFSet?"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setChanVoltageOffset(self, chNum, val):
        cstr = f":CHANnel{chNum}:OFFSet {val}"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getChanVoltsPerDiv(self, chNum):
        qstr = f":CHANnel{chNum}:SCALe?"
        rstr = self.query(qstr)
        return float(rstr)

    @BasicInstrument.tsdb_append
    def setChanVoltsPerDiv(self, chNum, vPerDiv):
        cstr = f":CHANnel{chNum}:SCALe {vPerDiv}"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getTriggerStatus(self):
        """Get Trigger Status.

        We expect TD, WAIT, RUN, AUTO, or STOP.
        """
        res = self.query(":TRIGger:STATus?")
        return res

    @BasicInstrument.tsdb_append
    def getRun(self):
        """Is acquisition running or stopped."""

        if self.getTriggerStatus() != "STOP":
            return True
        return False

    @BasicInstrument.tsdb_append
    def setRun(self, isRun):
        """Either enable run or stop the acquisition."""

        if isRun:
            self.run()
            return
        self.stop()

    @BasicInstrument.tsdb_append
    def getTimeBaseMode(self):
        """Get he mode of the horizontal timebase.

        We expect {MAIN|XY|ROLL}.
        MAIN stands for normal voltage vs time mode.
        XY stands X vs Y mode.
        ROLL stands for rolling mode.
        """
        res = self.query(":TIMebase:MODE?")
        return res

    @BasicInstrument.tsdb_append
    def setTimeBaseMode(self, val):
        """Set trigger mode.

        Takes {MAIN|XY|ROLL}.
        """
        self.write(f":TIMebase:MODE? {val}")

    @BasicInstrument.tsdb_append
    def getRoll(self):
        """Is Roll mode enabled."""
        if self.getTimeBaseMode() == "Roll":
            return True
        return False

    @BasicInstrument.tsdb_append
    def getTriggerMode(self):
        """Get trigger mode.

        We expect AUTO, NORM, or SING (for Single)
        """
        res = self.query(":TRIGger:SWEep?")
        return res

    @BasicInstrument.tsdb_append
    def setTriggerMode(self, val):
        """Set trigger mode.

        Takes AUTO, NORMal, or SINGle
        """
        self.write(f":TRIGger:SWEep {val}")

    @BasicInstrument.tsdb_append
    def getMemoryDepth(self):
        """Memory depth per channel.

        Returns
        -------
        Number corresponding to memory depth or AUTO
        """

        res = self.query(":ACQuire:MDEPth?")
        return res

    @BasicInstrument.tsdb_append
    def setMemoryDepth(self, val):
        """Set memory depth per channel.

        For 1 channel regime valid settings are
        {AUTO|12000|120000|1200000|12000000|24000000}
        Wherein, 24000000 (pts) is an optional memory depth.

        For 2 channel regime divide numbers by 2, i.e.
        {AUTO|6000|60000|600000|6000000|12000000}

        For 4 channel regime divide numbers by 4, i.e.
        {AUTO|3000|30000|300000|3000000|6000000}

        For 8 channel regime valid settings are
        {AUTO|12000|120000|1200000|12000000|24000000}.
        Wherein, 24000000 (pts) is an optional memory depth.

        For 16 channel regime valid settings are
        {AUTO|6000|60000| 600000|6000000|12000000}.
        Wherein, 12000000 (pts) is an optional memory depth.
        """

        # memory depth can be changed only when scope is not in STOP state
        running = self.getRun()
        if not running:
            self.setRun(True)
        self.write(f":ACQuire:MDEPth {val}")
        if not running:  # restore STOP/RUN state
            self.setRun(False)

    @BasicInstrument.tsdb_append
    def getSampleRate(self):
        """Get sample rate.

        Returns
        -------
        Sample rate in units of Samples/Second
        """

        res = self.query(":ACQuire:SRATe?")
        return float(res)

    def stop(self):
        self.write(":STOP")

    def run(self):
        self.write(":RUN")

    def restorePriorToFastGrab(self, chNum, old_config):
        """Restore relevant channel/scope settings prior to fast grab tune up."""
        self.setMemoryDepth(old_config["DeviceConfig"]["MemoryDepth"])

    def switchToFastGrab(self, chNum):
        """Switch scope to the fastest mode for trace delivery.

        To be fast, it should aim to decrease data transfer time,
        i.e. reduce number of transferred point.
        It also need to be interface dependent. GPIB and USB are known
        to be quite slow when compared to network connection.

        FIXME try to be smart about connection interface.

        Return
        ------
        old_config : dictionary
            old config with settings necessary to restore initial state
        """

        old_config = self.getConfig()
        self.setMemoryDepth(3000)

        return old_config

    def getRawWaveform(
        self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True
    ):
        """
        Get raw channel waveform in binary format.

        The idea is to have minimal processing from internal representation
        to actual volts, since it might require less queering of the scope
        or/and faster data transfer.

        Parameters
        ----------
        chNum : int
            Scope channel to use: 1, 2, 3, or 4
        availableNpnts : int or None (default)
            Available number of points. Do not set it if you want it auto detected.
        maxRequiredPoints : int
            Maximum number of required points, if we ask less than available
            we well get sparse set which proportionally fills all available time range.
        decimate : False or True (default)
            Decimate should be read as apply the low pass filter or not, technically
            for both setting we get decimation (i.e. smaller than available
            at the scope number of points). The name came from
            ``scipy.signal.decimate`` filtering function.
            If ``decimate=True`` is used, we get all available points
            and then low-pass filter them to get ``maxRequiredPoints``
            The result is less noisy then, but transfer time from the instrument
            is longer.
            If ``decimate=False``, then it we are skipping points to get needed number
            but we might see aliasing, if there is a high frequency noise
            and sparing > 1. Unless you know what you doing, it is recommended
            to use ``decimate=True``.
        """

        # if RAW is used the scope should be in STOP state
        self.write(f":WAVeform:SOURce CHAN{chNum}")
        self.write(
            ":WAVeform:MODE RAW"
        )  # {NORMal|MAXimum|RAW} RAW gives maximum number of points
        self.write(
            ":WAVeform:FORMat BYTE"
        )  # {WORD|BYTE|ASCii}, scope is 8 bit, BYTE is enough
        preamble = self.query(":WAVeform:PREamble?").split(",")
        """
        Format is
        <format>,<type>,<points>,<count>,<xincrement>,<xorigin>,<xreference>,<yincrement>,<yorigin>,<yreference>
        Wherein,
        <format>: 0 (BYTE), 1 (WORD) or 2 (ASC).
        <type>: 0 (NORMal), 1 (MAXimum) or 2 (RAW).
        <points>: an integer between 1 and 12000000. After the memory depth option is
                  installed, <points> is an integer between 1 and 24000000.
        <count>: the number of averages in the average sample mode
                 and 1 in other modes.
        <xincrement>: the time difference between two neighboring points
                      in the X direction.
        <xorigin>: the start time of the waveform data in the X direction.
        <xreference>: the reference time of the data point in the X direction.
        <yincrement>: the waveform increment in the Y direction.
        <yorigin>: the vertical offset relative
                   to the "Vertical Reference Position" in the Y direction.
        <yreference>: the vertical reference position in the Y direction.
        """
        rawChanCfg = {
            "format": int(preamble[0]),
            "type": int(preamble[1]),
            "availableNpnts": int(preamble[2]),
            "Navrg": int(preamble[3]),
            "xincrement": float(preamble[4]),
            "xorigin": float(preamble[5]),
            "xreference": int(preamble[6]),
            "yincrement": float(preamble[7]),
            "yorigin": int(preamble[8]),
            "yreference": int(preamble[9]),
        }
        logger.debug(f"rawChanCfg: {rawChanCfg}")
        availableNpnts = rawChanCfg["availableNpnts"]
        wfRaw = np.zeros(availableNpnts, dtype=np.uint8)
        maxreadable = 250_000  # the maximum number of bytes readable in one go
        chunk_size = 70_000  # unfortunately large chunk size prone to read errors
        errCnt = 0
        strt = 1
        stp = min(chunk_size, availableNpnts)
        errorFreeChunkSize = []
        errorProneChunkSize = []
        while strt <= availableNpnts:
            stp = strt - 1 + chunk_size
            stp = min(stp, availableNpnts)
            chunk_size = stp - strt + 1
            # reading requested number of points in chunks
            self.write(f":WAVeform:STARt {strt}")
            self.write(f":WAVeform:STOP {stp}")
            qstr = ":WAVeform:DATA?"
            try:
                wfRawChunk = self.query_binary_values(
                    qstr,
                    datatype="b",
                    header_fmt="ieee",
                    container=np.array,
                    chunk_size=(chunk_size + 100),
                )
                if len(wfRawChunk) == 0:
                    logger.info("Got empty chunk. Redoing.")
                    continue  # we need to repeat chunk read
                if len(wfRawChunk) != chunk_size:
                    logger.info(
                        "Expected chunk with length"
                        + f" {chunk_size} but got {len(wfRawChunk)}"
                    )
                    logger.info(
                        f"Current pointers are {strt=} {stp=} with {chunk_size=}"
                    )
                    logger.info("Redoing, chunk reading.")
                    continue  # we need to repeat chunk read
                wfRaw[strt - 1 : stp] = wfRawChunk  # noqa: E203 whitespace before ':'
                """
                All this craziness with tuning chunk_size
                and catching VisaIOError
                is because Rigol usbtmc connection is buggy.
                It present itself as high speed device over USB,
                but set incompatible packet size of 64
                while the USB standard dictates 512.
                In linux dmesg complains:
                'bulk endpoint 0x3 has invalid maxpacket 64'
                """
                strt += chunk_size
                errorFreeChunkSize.append(chunk_size)
                chunk_size = min(maxreadable, int(chunk_size * 1.1))
            except VisaIOError as err:
                logger.info(f"Detected recoverable {err}")
                errCnt += 1
                errorProneChunkSize.append(chunk_size)
                logger.debug(
                    f"Visa error count is {errCnt} while reading raw chunk the scope"
                )
                logger.debug(f"Current pointers are {strt=} {stp=} with {chunk_size=}")
                if len(errorFreeChunkSize) > 10:
                    chunk_size = int(np.mean(errorFreeChunkSize))
                else:
                    chunk_size = max(1, int(np.mean(errorProneChunkSize) * 0.8))
                logger.debug(f"New {chunk_size=}")
                logger.debug("Redoing, chunk reading.")
                pass  # we repeat this loop iteration again

        logger.debug(f"final {chunk_size=}")
        if maxRequiredPoints is None:
            maxRequiredPoints = self.maxRequiredPoints
        (
            sparsing,
            Npnts,
            availableNpnts,
            maxRequiredPoints,
        ) = calcSparsingAndNumPoints(
            availableNpnts=availableNpnts, maxRequiredPoints=maxRequiredPoints
        )
        rawChanCfg["Npnts"] = Npnts
        rawChanCfg["sparsing"] = sparsing
        if not decimate and sparsing > 1:
            wfRaw = wfRaw[::sparsing]

        trRaw = Trace(f"Ch{chNum}")
        trRaw.values = wfRaw.reshape(wfRaw.size, 1)
        if decimate and sparsing != 1:
            numtaps = 3
            # not sure it is the best case
            trRaw.values = scipy.signal.decimate(
                trRaw.values, sparsing, numtaps, axis=0
            )
        trRaw.config["unit"] = "Count"
        trRaw.config["tags"]["Decimate"] = decimate
        trRaw.config["tags"]["rawChanConfig"] = rawChanCfg
        return trRaw

    def getTimeTrace(self, rawChanCfg):
        timePerDiv = self.getTimePerDiv()
        trigDelay = self.getTrigDelay()
        availableNpnts = rawChanCfg["availableNpnts"]
        sparsing = rawChanCfg["sparsing"]
        Npnts = rawChanCfg["Npnts"]
        ind = np.linspace(0, Npnts - 1, Npnts)
        dx = rawChanCfg["xincrement"]
        xorig = rawChanCfg["xorigin"]
        tval = (ind - rawChanCfg["xreference"]) * dx * sparsing + xorig
        t = Trace("time")
        t.values = tval.reshape(tval.size, 1)
        t.config["unit"] = "S"
        t.config["tags"]["TimePerDiv"] = timePerDiv
        t.config["tags"]["TrigDelay"] = trigDelay
        t.config["tags"]["SampleRate"] = int(1 / dx)
        t.config["tags"]["AvailableNPnts"] = availableNpnts
        t.config["tags"]["Npnts"] = Npnts
        t.config["tags"]["Sparsing"] = sparsing
        return t

    def getWaveform(
        self, chNum, availableNpnts=None, maxRequiredPoints=None, decimate=True, **kwargs
    ):
        """
        For decimate use see ``getRawWaveform``.

        In short decimate=True is slower but more precise.
        """
        trRaw = self.getRawWaveform(
            chNum,
            availableNpnts=availableNpnts,
            maxRequiredPoints=maxRequiredPoints,
            decimate=decimate,
        )
        rawChanCfg = trRaw.config["tags"]["rawChanConfig"]
        VoltageOffset = self.getChanVoltageOffset(chNum)
        VoltsPerDiv = self.getChanVoltsPerDiv(chNum)
        tr = trRaw
        tr.values = (
            np.array(trRaw.values, dtype=int)
            - rawChanCfg["yreference"]
            - rawChanCfg["yorigin"]
        ) * rawChanCfg["yincrement"]

        tr.config["unit"] = "Volt"
        tr.config["tags"]["VoltageOffset"] = VoltageOffset
        tr.config["tags"]["VoltsPerDiv"] = VoltsPerDiv
        return tr


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    # instr = rm.open_resource("TCPIP::192.168.0.62::INSTR")
    instr = rm.open_resource("USB0::0x1AB1::0x04CE::DS1ZA170502787::0::INSTR")
    scope = RigolDS1054z(instr)
    print(f"ID: {scope.idn}")
    print(f"TimePerDiv = {scope.TimePerDiv}")
    # print(f'Ch1 mean: {scope.mean(1)}')
    print(f"Ch1 available points: {scope.getAvailableNumberOfPoints(1)}")
    print(f"Sample Rate: {scope.getSampleRate()}")
    print(f"Time per Div: {scope.getTimePerDiv()}")
    print(f"Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}")
    print(f"Ch1 Voltage Offset: {scope.getChanVoltageOffset(1)}")
    print("------ Header start -------------")
    print(str.join("\n", scope.getHeader()))
    print("------ Header ends  -------------")
    ch1 = scope.getTrace(1)
    traces = scope.getAllTraces()



import pytest
from qolab.data.trace import loadTrace
import numpy as np
import os


def isItExpectedTrace(tr):
    cfg = tr.getConfig()
    assert cfg["config"]["version"] == "0.1"
    assert cfg["config"]["model"] == "Trace"
    data = tr.getData()
    assert np.all((data - np.array([[1], [3], [2], [5]])) == 0)
    return True


def test_load_uncompressed_v0dot1_trace():
    tr = loadTrace("tests/trace_test_data/xtrace1.dat")
    assert isItExpectedTrace(tr) is True


def test_load_gzip_compressed_v0dot1_trace():
    tr = loadTrace("tests/trace_test_data/xtrace1.dat.gz")
    assert isItExpectedTrace(tr) is True


def test_load_bzip_compressed_v0dot1_trace():
    tr = loadTrace("tests/trace_test_data/xtrace1.dat.bz")
    assert isItExpectedTrace(tr) is True

    tr = loadTrace("tests/trace_test_data/xtrace1.dat.bz2")
    assert isItExpectedTrace(tr) is True


def test_tryCompressedIfMissing():
    fname = "tests/trace_test_data/only_compressed_file1.dat"
    # we check that the guess is working for gzip compressed file (.gz)
    tr = loadTrace(fname, tryCompressedIfMissing=True)
    assert isItExpectedTrace(tr) is True

    fname = "tests/trace_test_data/only_compressed_file2.dat"
    # we check that the guess is working for bzip compressed file (.bz)
    tr = loadTrace(fname, tryCompressedIfMissing=True)
    assert isItExpectedTrace(tr) is True

    fname = "tests/trace_test_data/only_compressed_file3.dat"
    # we check that the guess is working for bzip compressed file (.bz2)
    tr = loadTrace(fname, tryCompressedIfMissing=True)
    assert isItExpectedTrace(tr) is True

    # now we disable search for compressed version
    with pytest.raises(FileNotFoundError):
        tr = loadTrace(fname, tryCompressedIfMissing=False)

    tr = loadTrace("tests/trace_test_data/xtrace1.dat")
    assert isItExpectedTrace(tr) is True


def test_saving_raw():
    import tempfile

    with tempfile.NamedTemporaryFile(delete=True) as fp:
        fp.close()
        fname = fp.name
        assert not os.path.exists(fname)

        tr = loadTrace("tests/trace_test_data/xtrace1.dat")
        tr.save(fname, compressionmethod=None)
        trLoaded = loadTrace(fname)
        assert isItExpectedTrace(trLoaded)
        os.remove(fname)


def test_saving_gzip():
    import tempfile

    with tempfile.NamedTemporaryFile(delete=False) as fp:
        fp.close()
        fname = fp.name + ".bz"
        assert not os.path.exists(fname)

        tr = loadTrace("tests/trace_test_data/xtrace1.dat")
        tr.save(fname, compressionmethod="bzip")
        trLoaded = loadTrace(fname)
        assert isItExpectedTrace(trLoaded)
        os.remove(fname)



"""
Created by Eugeniy E. Mikhailov 2021/11/29
"""

from qolab.hardware.scope.sds1104x import SDS1104X
from qolab.hardware.scope._basic import calcSparsingAndNumPoints
from qolab.hardware.basic import BasicInstrument
from qolab.hardware.scpi import response2numStr
from qolab.data.trace import Trace
import numpy as np
import scipy.signal
from pyvisa.constants import InterfaceType
import platform

import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)8s %(name)s: %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class SDS800XHD(SDS1104X):
    """Siglent SDS800XHD scope"""

    # SDS1104x has actually 8 divisions but its behave like it has 10,
    # the grabbed trace has more points outside what is visible on the screen
    vertDivOnScreen = 8
    horizDivOnScreen = 10

    def __init__(self, resource, *args, **kwds):
        super().__init__(resource, *args, **kwds)
        self.config["Device model"] = "SDS800XHD"
        self.resource.read_termination = "\n"
        self.resource.timeout = 1000
        self.numberOfChannels = 4
        self.maxRequiredPoints = 1000
        # desired number of points per channel, can return twice more
        self.deviceProperties.update({"MemoryDepth"})

    @BasicInstrument.tsdb_append
    def getTimePerDiv(self):
        qstr = "TDIV?"
        rstr = self.query(qstr)
        # Careful! TDIV? is undocumented for SDS800XHD scope,
        # the prescribe command is ":TIMebase:SCALe?".
        # But "TDIV?"  works identical to SDS2304, i.e.
        # Siglent claims that this model should have same commands as SDS1104X
        # However response is different.
        # For example we got '2.00E-08S' instead 'TDIV 2.00E-08S'
        # expected reply to query: '2.00E-08S'
        prefix, numberString, unit = response2numStr(
            rstr, firstSeparator=None, unit="S"
        )
        return float(numberString)

    def getRawWaveform(
        self,
        chNum,
        availableNpnts=None,
        maxRequiredPoints=None,
        decimate=True,
        fmath=False,
    ):
        """
        Get raw channel waveform in binary format.

        Parameters
        ----------
        chNum : int
            Scope channel to use: 1, 2, 3, or 4
        availableNpnts : int or None (default)
            Available number of points. Do not set it if you want it auto detected.
        maxRequiredPoints : int
            Maximum number of required points, if we ask less than available
            we well get sparse set which proportionally fills all available time range.
        decimate : False or True (default)
            Decimate should be read as apply the low pass filter or not, technically
            for both setting we get decimation (i.e. smaller than available
            at the scope number of points). The name came from
            ``scipy.signal.decimate`` filtering function.
            If ``decimate=True`` is used, we get all available points
            and then low-pass filter them to get ``maxRequiredPoints``
            The result is less noisy then, but transfer time from the instrument
            is longer.
            If ``decimate=False``, then it we are skipping points to get needed number
            but we might see aliasing, if there is a high frequency noise
            and sparing > 1. Unless you know what you doing, it is recommended
            to use ``decimate=True``.
        """

        rawChanCfg = {}
        # switching to binary data transfer
        self.write(":WAVeform:WIDTh WORD")  # two bytes per data point
        rawChanCfg["WaveformWidth"] = "WORD"
        self.write(":WAVeform:BYTeorder LSB")
        rawChanCfg["WaveformByteorder"] = "LSB"

        if availableNpnts is None:
            # using channel 1 to get availableNpnts
            availableNpnts = self.getAvailableNumberOfPoints(1)
            rawChanCfg["availableNpnts"] = availableNpnts

        if maxRequiredPoints is None:
            maxRequiredPoints = self.maxRequiredPoints
        (
            sparsing,
            Npnts,
            availableNpnts,
            maxRequiredPoints,
        ) = calcSparsingAndNumPoints(availableNpnts, maxRequiredPoints)
        rawChanCfg["Npnts"] = Npnts
        rawChanCfg["sparsing"] = sparsing
        if decimate:
            Npnts = availableNpnts  # get all of them and decimate later
        if (sparsing == 1 and Npnts == availableNpnts) or decimate:
            # We are getting all points of the trace
            self.write(":WAVeform:STARt 0")  # start point to read from the scope memory
            self.write(f":WAVeform:MAXPoint {availableNpnts}")  # last point to read
            self.write(":WAVeform:INTerval 1")  # sparsing of 1, i.e. read every point
            self.write(f":WAVeform:POINt {availableNpnts}")  # transfer all points
        else:
            # we just ask every point with 'sparsing' interval
            # fast to grab but we could do better with more advance decimate
            # method, which allow better precision for the price
            # of longer acquisition time
            self.write(":WAVeform:STARt 0")  # start point to read from the scope memory
            self.write(f":WAVeform:MAXPoint {availableNpnts}")  # last point to read
            self.write(f":WAVeform:INTerval {sparsing}")  #  interval between points
            # Note: it is not enough to provide sparsing
            # number of requested points needed to be asked too.
            # However this scope is smart enough to truncate the output to
            # physically available points, if you request more no harm is done.
            self.write(f":WAVeform:POINt {Npnts}")  # transfer all points

        trRaw = Trace(f"Ch{chNum}")
        if fmath:
            self.write(f":WAVeform:SOURce F{chNum}")
        else:
            self.write(f":WAVeform:SOURce C{chNum}")
        qstr = ":WAVeform:DATA?"
        if self.resource.interface_type == InterfaceType.usb:
            # Setting chunk size to 496 bytes, it seems that SDS sends data
            # in 512 bytes chunks via USB.
            # Which is 8 packets of 64 bytes, but each packet takes 2 bytes for a header.
            # Thus useful payload is 512-8*2 = 496
            # see https://patchwork.ozlabs.org/project/qemu-devel/patch/20200317095049.28486-4-kraxel@redhat.com/
            # Setting chunk_size for a large number has *catastrophic* results
            # on data transfer rate, since we wait for more data
            # which is not going to come until timeout expires
            # Setting it low is not as bad but still slows down the transfer.
            # NOTE: I am not sure if it is a Linux driver issue or more global.
            # The transfer rate is about
            #       5550 kB/S, for 10k points
            #       1400 kB/S, for 50k points
            #       1000 kB/S, for 100k points
            #       500  kB/S, for 500k points
            #       160  kB/S, for 1000k points
            #       55   kB/S, for 2.5M points
            # It is about factor of 2 slower (for 100k points),
            # if the scope is in the Run mode, i.e. not Stopped.
            # FIXME find why speed depends on number of points.
            wfRaw = self.query_binary_values(
                qstr,
                datatype="h",
                header_fmt="ieee",
                container=np.array,
                chunk_size=496,
            )
            if "Windows" == (platform.system()):
                # somehow on windows there is an extra '\n' at the end
                _ = self.read_bytes(1)
        else:
            wfRaw = self.query_binary_values(
                qstr, datatype="h", header_fmt="ieee", container=np.array
            )
        trRaw.values = wfRaw.reshape(wfRaw.size, 1)
        if decimate and sparsing != 1:
            numtaps = 3
            # not sure it is the best case
            trRaw.values = scipy.signal.decimate(
                trRaw.values, sparsing, numtaps, axis=0
            )

        trRaw.config["unit"] = "Count"
        trRaw.config["tags"]["Decimate"] = decimate
        trRaw.config["tags"]["rawChanConfig"] = rawChanCfg
        return trRaw

    def getParsedPreamble(self):
        """
        Parse preamble generated by the scope.
        It has a lot of information, number of points, scale, bits resolution, etc.

        Note: preamble is channel specific! The prior code must set desired channel
        with self.write(":WAVeform:SOURce C{chNum}")
        """
        preamble = {}
        qstr = ":WAVeform:PREamble?"
        if self.resource.interface_type == InterfaceType.usb:
            # Setting chunk size to 496 bytes, it seems that SDS sends data
            # in 512 bytes chunks via USB.
            # Which is 8 packets of 64 bytes, but each packet takes 2 bytes for a header.
            # Thus useful payload is 512-8*2 = 496
            # see https://patchwork.ozlabs.org/project/qemu-devel/patch/20200317095049.28486-4-kraxel@redhat.com/
            # Setting chunk_size for a large number has *catastrophic* results
            # on data transfer rate, since we wait for more data
            # which is not going to come until timeout expires
            resp_bin = self.query_binary_values(
                qstr,
                datatype="c",
                header_fmt="ieee",
                container=np.array,
                chunk_size=496,
            )
        else:
            resp_bin = self.query_binary_values(
                qstr, datatype="c", header_fmt="ieee", container=np.array
            )
        preamble["Npoints"] = resp_bin[116:120].view(np.int32).item()
        preamble["firstPoint"] = resp_bin[132:136].view(np.int32).item()
        preamble["sparsing"] = resp_bin[136:140].view(np.int32).item()
        preamble["voltsPerDiv"] = resp_bin[156:160].view(np.float32).item()
        preamble["verticalOffset"] = resp_bin[160:164].view(np.float32).item()
        preamble["codePerDiv"] = resp_bin[164:168].view(np.float32).item()
        preamble["adcBit"] = resp_bin[172:174].view(np.int16).item()
        preamble["samplingTime"] = resp_bin[176:180].view(np.float32).item()
        preamble["trigDelay"] = (
            -resp_bin[180:188].view(np.float64).item()
        )  # manual is wrong, it claims that this int64

        return preamble

    def getWaveform(
        self,
        chNum,
        availableNpnts=None,
        maxRequiredPoints=None,
        decimate=True,
        fmath=False,
    ):
        """
        For decimate use see ``getRawWaveform``.

        In short decimate=True is slower but more precise.
        """
        trRaw = self.getRawWaveform(
            chNum,
            availableNpnts=availableNpnts,
            maxRequiredPoints=maxRequiredPoints,
            decimate=decimate,
            fmath=fmath,
        )
        preamble = self.getParsedPreamble()
        VoltageOffset = preamble["verticalOffset"]
        VoltsPerDiv = preamble["voltsPerDiv"]
        tr = trRaw
        tr.values = trRaw.values * VoltsPerDiv / preamble["codePerDiv"] - VoltageOffset
        tr.config["unit"] = "Volt"
        tr.config["tags"]["VoltageOffset"] = VoltageOffset
        tr.config["tags"]["VoltsPerDiv"] = VoltsPerDiv
        tr.config["tags"]["Preamble"] = preamble
        return tr

    @BasicInstrument.tsdb_append
    def setSampleRate(self, val):
        """
        Set scope sampling rate

        Note: Memory management should be set to fixed
        sampling rate otherwise this command has no effect,
        while reporting success.
        """
        self.write(
            ":ACQuire:MMANagement FSRate"
        )  # switch to fixed sampling rate setting
        cstr = f":ACQuire:SRATe {val}"
        self.write(cstr)

    @BasicInstrument.tsdb_append
    def getMemoryDepth(self):
        rstr = self.query(":ACQuire:MDEPth?")
        if rstr[-1] == "G":
            return int(rstr[:-1]) * 1_000_000_000
        if rstr[-1] == "M":
            return int(rstr[:-1]) * 1_000_000
        if rstr[-1] == "k":
            return int(rstr[:-1]) * 1_000
        return int(rstr)

    @BasicInstrument.tsdb_append
    def setMemoryDepth(self, val):
        """
        Set scope memory depth. Only predefined values are possible.

        Note: Memory management should be set to fixed
        memory depth otherwise this command has no effect,
        while reporting success.
        """
        self.write(
            ":ACQuire:MMANagement FMDepth"
        )  # switch to fixed memory depth setting
        # Note: 1k setting is not possible, but manual claims otherwise
        # if val <= 1e3:
        # depth = "1k"
        if val <= 10e3:
            depth = "10k"
        elif val <= 100e3:
            depth = "100k"
        elif val <= 1e6:
            depth = "1M"
        elif val <= 10e6:
            depth = "10M"
        else:
            logger.info(
                "Memory depth higher than 10M are possible for 2 or 1 channels acquisition modes."
            )
            logger.info("If you know what you doing set it manually.")
            logger.info("For now are setting to 10M memory depth.")
            depth = "10M"
            # Note: the manual states that 100M possible
            # but I do not see it in the front panel
            # here are the values for achievable settings
            # depth = "25M"   # possible for 2 channels
            # depth = "50M"   # possible for 1 channels

        cstr = f":ACQuire:MDEPth {depth}"
        self.write(cstr)


if __name__ == "__main__":
    import pyvisa

    print("testing")
    rm = pyvisa.ResourceManager()
    print(rm.list_resources())
    # instr = rm.open_resource("TCPIP::192.168.0.62::INSTR")
    instr = rm.open_resource("USB0::62700::4119::SDS08A0X806445::0::INSTR")
    scope = SDS800XHD(instr)
    print(f"ID: {scope.idn}")
    # print(f'Ch1 mean: {scope.mean(1)}')
    print(f"Ch1 available points: {scope.getAvailableNumberOfPoints(1)}")
    print(f"Sample Rate: {scope.getSampleRate()}")
    print(f"Time per Div: {scope.getTimePerDiv()}")
    print(f"Ch1 Volts per Div: {scope.getChanVoltsPerDiv(1)}")
    print(f"Ch1 Voltage Offset: {scope.getChanVoltageOffset(1)}")
    print("------ Header start -------------")
    print(str.join("\n", scope.getHeader()))
    print("------ Header ends  -------------")
    ch1 = scope.getTrace(1)
    # traces = scope.getAllTraces()



#!/usr/bin/env python
# encoding: utf-8

"""
rubidium.py

Software suite for calculating Rubidium absorption spectra.

Example usage
    T = 273.15 + 35 # Temperature in Kelvin
    Lc = 0.075 # Length of cell in meters
    delta = linspace(-4,6,200)  # set up detuning values
    absdata = AbsorptionProfile(delta*1e9,T,Lc)  # calculate absorption
    plot(delta, absdata)  # plot absorption data

This code is based on the paper by
Siddons et al. J. Phys. B: At. Mol. Opt. Phys. 41, 155004 (2008).

And heavily inspired by Mathematica code by the same author:
http://massey.dur.ac.uk/resources/resources.html
http://massey.dur.ac.uk/resources/psiddons/absdisD2.nb

Created by Andrew M. C. Dawes on 2015-07-15.
Copyright (c) 2015 Andrew Dawes. Some rights reserved.

This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


from scipy.constants import hbar, epsilon_0, c, m_u, k, m_e, alpha
from scipy.special import erf
from scipy.interpolate import interp1d
from numpy import sqrt, pi, exp, real, power
from numpy import nan_to_num, roll, seterr
import numpy as np

# ignoring invalid warnings as the erf is touchy near ∆=0.
# No interesting physics there so we can safely ignore it.
seterr(invalid="ignore")

a_0 = hbar / (m_e * c * alpha)  # bohr radius


def dispersion(a, y):
    """Calculate dispersion profile for parameter a and frequency y"""
    disp = nan_to_num(
        1j
        / 2.0
        * sqrt(pi)
        * (
            exp(1 / 4.0 * (a - 2j * y) ** 2) * (1 - erf(a / 2 - 1j * y))
            - exp(1 / 4.0 * (a + 2 * 1j * y) ** 2) * (1 - erf(a / 2.0 + 1j * y))
        )
    )
    return disp


def voigt(a, y):
    """Calculate Voigt profile for parameter a and frequency y"""
    v = nan_to_num(
        sqrt(pi)
        / 2.0
        * (
            exp(1 / 4.0 * (a - 2j * y) ** 2)
            * (
                (1 - erf(a / 2.0 - 1j * y))
                + exp(2 * 1j * a * y) * (1 - erf(a / 2.0 + 1j * y))
            )
        )
    )
    return v


class Rubidium:
    """
    Rubidium class.

    Attributes:
    :param int line: Set the D-line. 1: D1, 2: D2. Defaults to 1.
    :param int abundance: Abundance of Rb-85. Defaults to 0.7217 (natural abundance).
        Abundance of Rb-87 is 1-abundance.

    """

    def __init__(self, line=1, abundance=0.7217):
        # === private ===
        self._lProbe = 794.978e-9
        self._kProbe = 2 * pi / self._lProbe
        self._omegaProbe = self._kProbe * c
        self._Gamma2 = 2 * pi * 5.746e6  # Excited state decay rate
        self._d21 = sqrt(3) * sqrt(
            3 * epsilon_0 * hbar * self._Gamma2 * self._lProbe**3 / (8 * pi**2)
        )  # reduced matrix element

        # Transition strength factors C_F^2, call as F87[1][2] for the F=1 to F'=2 strength
        self._F87 = None
        self._F85 = None

        # Detuning factors, call as det87[1][2] for the F=1 to F'=2 hyperfine transition.
        self._detF87 = None
        self._detF85 = None

        # === public (see setter) ===
        self._line = self.line = line
        self._abundance = self.abundance = abundance

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, val):
        if val == 1 or val == 2:
            self._line = val
        else:
            raise ValueError("Line can only be 1 (D1) or 2 (D2).")

        if val == 1:
            self._lProbe = 794.978e-9
            self._Gamma2 = 2 * pi * 5.746e6
            self._F87 = [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 1 / 18.0, 5 / 18.0, 0.0],
                [0.0, 5 / 18.0, 5 / 18.0, 0.0],
            ]
            self._F85 = [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 10 / 81.0, 35 / 81.0, 0.0],
                [0.0, 0.0, 35 / 81.0, 28 / 81.0, 0.0],
            ]

            # Detuning factors, call as det87[1][2] for the F=1 to F'=2 hyperfine transition,
            # note the sign differs from Siddon's paper. Not sure why I did that? FIXME?
            self._detF87 = [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, -3820.046e6, -4632.339e6, 0.0],
                [0.0, 3014.644e6, 2202.381e6, 0.0],
            ]

            self._detF85 = [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, -1538.063e6, -1900.087e6, 0.0],
                [0.0, 0.0, 1497.657e6, 1135.721e6, 0.0],
            ]

        if val == 2:
            self._lProbe = 780.241e-9
            self._Gamma2 = 2 * pi * 6.065e6
            self._F87 = [
                [0.0, 0.0, 0.0, 0.0],
                [1 / 9.0, 5 / 18.0, 5 / 18.0, 0.0],
                [0.0, 1 / 18.0, 5 / 18.0, 7 / 9.0],
            ]
            self._F85 = [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1 / 3.0, 35 / 81.0, 28 / 81.0, 0.0],
                [0.0, 0.0, 10 / 81.0, 35 / 81.0, 1.0],
            ]
            self._detF87 = [
                [0.0, 0.0, 0.0, 0.0],
                [-4027.403e6, -4099.625e6, -4256.570e6, 0.0],
                [0.0, 2735.050e6, 2578.110e6, 2311.260e6],
            ]

            self._detF85 = [
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, -1635.454e6, -1664.714e6, -1728.134e6, 0.0],
                [0.0, 0.0, 1371.290e6, 1307.870e6, 1186.910e6],
            ]

        self._kProbe = 2 * pi / self._lProbe
        self._omegaProbe = self._kProbe * c
        self._d21 = sqrt(3) * sqrt(
            3 * epsilon_0 * hbar * self._Gamma2 * self._lProbe**3 / (8 * pi**2)
        )

    @property
    def abundance(self):
        return self._abundance

    @abundance.setter
    def abundance(self, val):
        if 0 <= val <= 1:
            self._abundance = val
        else:
            raise ValueError("Abundance must be between 0 and 1.")

    @staticmethod
    def u87(T):
        """
        Mean thermal velocity for 87-Rb

        :param float T: Temperature in K
        :return: Mean thermal velocity
        """
        return sqrt(2 * k * T / (86.909180520 * m_u))

    @staticmethod
    def u85(T):
        """
        Mean thermal velocity for 85-Rb

        :param float T: Temperature in K
        :return: Mean thermal velocity
        """
        return sqrt(2 * k * T / (84.911789732 * m_u))

    @staticmethod
    def P(T):
        """
        Vapor pressure in a thermal cell

        :param float T: Temperature in K
        :return: Vapor pressure in Pa
        """
        torrToPa = 133.3223684210526  # 101325/760 Pa/Torr,  1 atmosphere is 760 torr
        # These numbers are from Steck Rb  87 D line data section 2
        # https://steck.us/alkalidata/ (checked in 2025) but it seems to use reference from 1984
        # pressure will be within 5% in interval 298K-550K according to the reference
        c1 = 2.881  # if it zero formula would report pressure in Atmospheres
        if T > 312.46:  # phase transition is at 39.31C
            # liquid phase
            c2 = 4.312
            c3 = 4040
        else:
            # solid phase
            c2 = 4.857
            c3 = 4215
        pInTorrs = power(10, (c1 + c2 - c3 / T))
        return torrToPa * pInTorrs

    def N(self, T, isotope):
        """
        Atomic number density at T for given isotope
        :param float T: Temperature in K
        :param str isotope: either '85' or '87' or 'all' for total density
        :return: Density (in 1/m**3)
        """
        if isotope == "85":
            abundance = self.abundance
        elif isotope == "87":
            abundance = 1 - self.abundance
        elif isotope == "all":
            abundance = 1
        else:
            raise ValueError("unknown isotope value {isotope}")

        return abundance * self.P(T) / (k * T)

    def lo87(self, T):
        """
        This is used as the parameter a in the voigt and dispersion calculations

        :param float T: Temperature in K
        """
        return self._Gamma2 / (self._kProbe * self.u87(T))

    def D87(self, y, T):
        """
        Calculate the dispersion for Rb-87 at frequency y and temperature T

        :param float y: Frequency
        :param float T: Temperature in K
        """
        return dispersion(self.lo87(T), y)

    def V87(self, y, T):
        """
        Calculate the voigt profile for Rb-87 at frequency y and temperature T
        :param float y: Frequency
        :param float T: Temperature in K
        """
        return voigt(self.lo87(T), y)

    def lo85(self, T):
        """This is used as the parameter a in the voigt and dispersion calculations"""
        return self._Gamma2 / (self._kProbe * self.u85(T))

    def D85(self, y, T):
        """
        Calculate the dispersion for Rb-85 at frequency y and temperature T
        :param float y: Frequency
        :param float T: Temperature in K
        """
        return dispersion(self.lo85(T), y)

    def V85(self, y, T):
        """
        Calculate the voigt profile for Rb-85 at frequency y and temperature T
        :param float y: Frequency
        :param float T: Temperature in K
        """
        return voigt(self.lo85(T), y)

    def K87(self, T, Fg, Fe):
        """

        :param T: Temperature in K
        :param Fg: Hyperfine quantum number of ground state
        :param Fe: Hyperfine quantum number of excited state
        :return:
        """
        return (
            self._F87[Fg][Fe]
            * 1
            / 8.0
            * 1
            / (hbar * epsilon_0)
            * self._d21**2
            * self.N(T, "87")
            / (self._kProbe * self.u87(T))
        )

    def K85(self, T, Fg, Fe):
        """

        :param T: Temperature in K
        :param Fg: Hyperfine quantum number of ground state
        :param Fe: Hyperfine quantum number of excited state
        :return:
        """
        return (
            self._F85[Fg][Fe]
            * 1
            / 12.0
            * 1
            / (hbar * epsilon_0)
            * self._d21**2
            * self.N(T, "85")
            / (self._kProbe * self.u85(T))
        )

    def chiRe87(self, delta, T, Fg, Fe):
        """
        The real part of the susceptibility for 87-Rb

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :param int Fg: Hyperfine quantum number of ground state
        :param int Fe: Hyperfine quantum number of excited state
        :return: The real part of the susceptibility for 87-Rb
        """
        return self.K87(T, Fg, Fe) * real(
            self.D87(
                2 * pi * (delta + self._detF87[Fg][Fe]) / (self._kProbe * self.u87(T)),
                T,
            )
        )

    def chiIm87(self, delta, T, Fg, Fe):
        """
        The imaginary part of the susceptibility for 87-Rb

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :param int Fg: Hyperfine quantum number of ground state
        :param int Fe: Hyperfine quantum number of excited state
        :return: The imaginary part of the susceptibility for 87-Rb
        """
        return self.K87(T, Fg, Fe) * real(
            self.V87(
                2 * pi * (delta + self._detF87[Fg][Fe]) / (self._kProbe * self.u87(T)),
                T,
            )
        )

    def chiRe85(self, delta, T, Fg, Fe):
        """
        The real part of the susceptibility for 85-Rb

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :param int Fg: Hyperfine quantum number of ground state
        :param int Fe: Hyperfine quantum number of excited state
        :return: The real part of the susceptibility for 85-Rb
        """
        return self.K85(T, Fg, Fe) * real(
            self.D85(
                2 * pi * (delta + self._detF85[Fg][Fe]) / (self._kProbe * self.u85(T)),
                T,
            )
        )

    def chiIm85(self, delta, T, Fg, Fe):
        """
        The imaginary part of the susceptibility for 85-Rb

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :param int Fg: Hyperfine quantum number of ground state
        :param int Fe: Hyperfine quantum number of excited state
        :return: The imaginary part of the susceptibility for 85-Rb
        """
        return self.K85(T, Fg, Fe) * real(
            self.V85(
                2 * pi * (delta + self._detF85[Fg][Fe]) / (self._kProbe * self.u85(T)),
                T,
            )
        )

    def TotalChi(self, delta, T):
        """
        The total real and imaginary part of the susceptibility is the sum of each transition

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :return: total real part of the susceptibility, total imaginary part of the susceptibility

        """
        total_real = 0
        total_imag = 0

        if self.line == 1:  # D1 line
            for idx_g in range(1, 3):  # sum over all Rb-87 transitions (Fg=1,2; Fe=1,2)
                for idx_e in range(1, 3):
                    total_real += self.chiRe87(delta, T, idx_g, idx_e)
                    total_imag += self.chiIm87(delta, T, idx_g, idx_e)

            for idx_g in range(2, 4):  # sum over all Rb-85 transitions (Fg=2,3; Fe=2,3)
                for idx_e in range(2, 4):
                    total_real += self.chiRe85(delta, T, idx_g, idx_e)
                    total_imag += self.chiIm85(delta, T, idx_g, idx_e)

        elif self.line == 2:  # D2 line
            for idx_g in range(
                1, 3
            ):  # sum over all Rb-87 transitions (Fg=1,2; Fe=0,1,2,3; Delta_F=1)
                for idx_e in range(-1, 2):
                    total_real += self.chiRe87(delta, T, idx_g, idx_e + idx_g)
                    total_imag += self.chiIm87(delta, T, idx_g, idx_e + idx_g)

            for idx_g in range(
                2, 4
            ):  # sum over all Rb-85 transitions (Fg=2,3; Fe=1,2,3,4; Delta_F=1)
                for idx_e in range(-1, 2):
                    total_real += self.chiRe85(delta, T, idx_g, idx_e + idx_g)
                    total_imag += self.chiIm85(delta, T, idx_g, idx_e + idx_g)

        return total_real, total_imag

    def Totaln(self, delta, T):
        """
        Calculate refractive index

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :return: Refractive index
        """
        total_chi_real, _ = self.TotalChi(delta, T)
        return 1 + 1 / 2.0 * total_chi_real

    def TotalAlpha(self, delta, T):
        """
        Calculate absorption coefficient

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :return: Absorption coefficient
        """
        _, total_chi_imag = self.TotalChi(delta, T)
        return self._kProbe * total_chi_imag

    def getTotalAphaPerAtomFn(self, T_typical=333):
        """
        Returns function to calculate absorption spectrum per atom based on interpolation.

        It recalculate alpha for a set of detuning,
        and returns interpolation function. Which is much faster for bulk calculations.

        Unlike in TotalAlpha the result is divided by total number
        of atoms. Which makes it handy for fitting atomic transmission
        in a search for density.

        T_typical - temperature in K used to calculate absorption
        Note that shape of absorption weakly depends on temperature
        since thermal velocity (u_T) changes too, but it is weak dependence for a typical
        hot cell temperatures: u_T ~ sqrt(T) i.e. about 5% in the range 20C-100C

        Example:
        > abs_per_atom = getTotalAphaPerAtomFn(T_typical = 333)
        > det = np.linspace(-5, 7, 10000) * 1e9  # detunings in Hz
        > spectrum = abs_per_atom(det)
        > plt.plot(det, spectrum)

        """
        detunings = np.linspace(-5, 7, 1000) * 1e9  # detunings in Hz
        abs_coeff = self.TotalAlpha(detunings, T_typical) / self.N(T_typical, "all")
        # pad range of detuning to get 0 for detunings far away from resonances
        detunings = np.concatenate(([-5e30], detunings, [5e30]))
        abs_coeff = np.concatenate(([0], abs_coeff, [0]))

        return interp1d(detunings, abs_coeff)

    def Transmission(self, delta, T, Lc):
        """
        Transmission as a function of detuning, temperature, and cell length

        :param delta: Relative frequency detuning
        :param T: Temperature in K
        :param Lc: Cell length
        :return: Transmission through cell
        """
        return exp(-self.TotalAlpha(delta, T) * Lc)

    def groupVelocity(self, delta, T):
        """

        :param float delta: Relative frequency detuning
        :param float T: Temperature in K
        :return: Group velocity
        """
        ndata = self.Totaln(delta * 1e9, T)
        dndw = (roll(ndata, -1) - ndata) / (
            (delta[1] - delta[0]) * 1e9
        )  # numerical derivative dn/dw
        ng = (
            ndata[:-1] + (delta[:-1] * 1e9 + self._omegaProbe) * dndw[:-1]
        )  # group index
        vg = 3e8 / ng
        return vg

    def transition_frequency87(self, Fg, Fe, Fe2=None):
        """
        Return the absolute frequency of the hyperfine transition Fg -> Fe for Rb87.

        Numbers from http://steck.us/alkalidata/

        :param Fg: Hyperfine quantum number F of the ground state (D1: 1,2; D2: 1,2)
        :param Fe: Hyperfine quantum number F of the excited state (D1: 1,2; D2: 0,1,2,3)
        :param Fe2: Optional parameter. If given, the cross-over transition (Fe, Fe2) is computed.
        :return: Transition frequency in Hz
        """
        freq = None

        if self.line == 1:  # D1 line
            # consistency check
            if not (isinstance(Fg, int) and isinstance(Fe, int)):
                raise TypeError

            if not ((1 <= Fg <= 2) and (1 <= Fe <= 2)):
                raise ValueError

            if Fe2 is not None:
                if not isinstance(Fe2, int):
                    raise TypeError
                if not (1 <= Fe2 <= 2):
                    raise ValueError

            freq = 377.107463380e12  # (S->P transition of Rb87 D1 line)

            freq_g = [0, 4.271676631815181e9, -2.563005979089109e9]
            freq_e = [0, -509.06e6, 305.44e6]

            if Fe2 is None:
                freq += freq_g[Fg] + freq_e[Fe]
            else:
                freq += freq_g[Fg] + (freq_e[Fe] + freq_e[Fe2]) / 2

        if self.line == 2:  # D2 line
            # consistency check
            if not (isinstance(Fg, int) and isinstance(Fe, int)):
                raise TypeError

            if not ((1 <= Fg <= 2) and (0 <= Fe <= 3)):
                raise ValueError

            if Fe2 is not None:
                if not isinstance(Fe2, int):
                    raise TypeError
                if not (0 <= Fe2 <= 3):
                    raise ValueError

            freq = 384.2304844685e12  # (S->P transition of Rb87 D2 line)

            freq_g = [0, 4.271676631815181e9, -2.563005979089109e9]
            freq_e = [-302.0738e6, -229.8518e6, -72.9112e6, 193.7407e6]

            if Fe2 is None:
                freq += freq_g[Fg] + freq_e[Fe]
            else:
                freq += freq_g[Fg] + (freq_e[Fe] + freq_e[Fe2]) / 2

        return freq

    def transition_frequency85(self, Fg, Fe, Fe2=None):
        """
        Return the absolute frequency of the hyperfine transition Fg -> Fe for Rb85.

        Numbers from http://steck.us/alkalidata/

        :param Fg: Hyperfine quantum number F of the ground state (D1: 2,3; D2: 2,3)
        :param Fe: Hyperfine quantum number F of the excited state (D1: 2,3; 1,2,3,4)
        :param Fe2: Optional parameter. If given, the cross-over transition (Fe, Fe2) is computed.
        :return: Transition frequency in Hz
        """
        freq = None

        if self.line == 1:  # D1 line
            # consistency check
            if not (isinstance(Fg, int) and isinstance(Fe, int)):
                raise TypeError

            if not ((2 <= Fg <= 3) and (2 <= Fe <= 3)):
                raise ValueError

            if Fe2 is not None:
                if not isinstance(Fe2, int):
                    raise TypeError
                if not (2 <= Fe2 <= 3):
                    raise ValueError

            freq = 377.107385690e12  # (S->P transition of Rb85 D1 line)

            freq_g = [0, 0, 1.7708439228e9, -1.2648885163e9]
            freq_e = [0, 0, -210.923e6, 150.659e6]

            if Fe2 is None:
                freq += freq_g[Fg] + freq_e[Fe]
            else:
                freq += freq_g[Fg] + (freq_e[Fe] + freq_e[Fe2]) / 2

        if self.line == 2:  # D2 line
            # consistency check
            if not (isinstance(Fg, int) and isinstance(Fe, int)):
                raise TypeError

            if not ((2 <= Fg <= 3) and (1 <= Fe <= 4)):
                raise ValueError

            if Fe2 is not None:
                if not isinstance(Fe2, int):
                    raise TypeError
                if not (1 <= Fe2 <= 4):
                    raise ValueError

            freq = 384.230406373e12  # (S->P transition of Rb85 D2 line)

            freq_g = [0, 0, 1.7708439228e9, -1.2648885163e9]
            freq_e = [0, -113.208e6, -83.835e6, -20.435e6, 100.205e6]

            if Fe2 is None:
                freq += freq_g[Fg] + freq_e[Fe]
            else:
                freq += freq_g[Fg] + (freq_e[Fe] + freq_e[Fe2]) / 2

        return freq


if __name__ == "__main__":
    from pylab import plot, show, linspace, title, xlabel, ylabel, subplot

    # ==================== D1 =====================
    rb = Rubidium(line=1)
    T = 273.15 + 30  # Temperature in Kelvin
    Lc = 0.075  # Length of cell in meters
    delta = linspace(-4, 6, 200)  # detuning in GHz
    transdata = rb.Transmission(delta * 1e9, T, Lc)
    ndata = rb.Totaln(delta * 1e9, T)

    vg = rb.groupVelocity(delta, T)
    transit = Lc / vg
    timeshift = transit - Lc / c

    subplot(3, 1, 1)
    plot(delta, transdata)
    title("Rubidium D1 Spectrum at T= " + str(T) + " K")
    ylabel("Transmission (Arb. Units)")
    subplot(3, 1, 2)
    plot(delta, ndata)
    ylabel("Index n")
    subplot(3, 1, 3)
    plot(delta[:-1], timeshift * 1e9)
    plot([delta[0], delta[-2]], [0, 0], color="#888888", linestyle="--", linewidth=1)
    xlabel(r"Detuning $\Delta$ (GHz)")
    ylabel(r"T (ns)")
    show()

    # ==================== D2 =====================
    rb = Rubidium(line=2)
    T = 273.15 + 47  # Temperature in Kelvin
    Lc = 0.075  # Length of cell in meters
    delta = linspace(-4, 6, 200)  # detuning in GHz
    transdata = rb.Transmission(delta * 1e9, T, Lc)
    ndata = rb.Totaln(delta * 1e9, T)

    vg = rb.groupVelocity(delta, T)
    transit = Lc / vg
    timeshift = transit - Lc / c

    subplot(3, 1, 1)
    plot(delta, transdata)
    title("Rubidium D2 Spectrum at T= " + str(T) + " K")
    ylabel("Transmission (Arb. Units)")
    subplot(3, 1, 2)
    plot(delta, ndata)
    ylabel("Index n")
    subplot(3, 1, 3)
    plot(delta[:-1], timeshift * 1e9)
    plot([delta[0], delta[-2]], [0, 0], color="#888888", linestyle="--", linewidth=1)
    xlabel(r"Detuning $\Delta$ (GHz)")
    ylabel(r"T (ns)")
    show()



"""Basic DAQ to be included into hardware aware classes."""

from qolab.hardware.basic import BasicInstrument


class DAQ(BasicInstrument):
    # Minimal set of methods to be implemented by a RFGenerator
    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "DAQ"
        self.config["Device model"] = "Generic DAQ Without Hardware interface"
        self.config["FnamePrefix"] = "daq"
        self.deviceProperties.update({"AnalogInputsNum", "AnalogOutputsNum"})

        # this is device dependent
        self.AnalogInputsNum = 0
        self.AnalogOutputsNum = 0



"""
Provide basic classes to operate scopes
"""

from .sds1104x import SDS1104X
from .sds2304x import SDS2304X
from .sds800xhd import SDS800XHD
from .rigolds1054z import RigolDS1054z

__all__ = ["SDS1104X", "SDS2304X", "SDS800XHD", "RigolDS1054z"]



"""Basic RF generator classes to implement hardware aware classes."""

from qolab.hardware.scpi import SCPIinstr
from qolab.hardware.basic import BasicInstrument


class RFGenerator(BasicInstrument):
    """Minimal set of methods to be implemented by a RF Generator.

    Intended to be used as a parent for hardware aware classes.
    """

    def __init__(self, *args, **kwds):
        BasicInstrument.__init__(self, *args, **kwds)
        self.config["Device type"] = "RFGenerator"
        self.config["Device model"] = "Generic RF generator Without Hardware interface"
        self.config["FnamePrefix"] = "rfgen"
        self.deviceProperties.update({"FreqFixed"})


class RFGeneratorSCPI(SCPIinstr, RFGenerator):
    """SCPI aware RF generator.

    Example
    -------

    >>> rm = pyvisa.ResourceManager()
    >>> RFGeneratorSCPI(rm.open_resource('TCPIP::192.168.0.2::INSTR'))
    """

    def __init__(self, resource, *args, **kwds):
        SCPIinstr.__init__(self, resource)
        RFGenerator.__init__(self, *args, **kwds)
        self.config["DeviceId"] = str.strip(self.idn)



from qolab.hardware.basic import BasicInstrument
from ._basic import RFGenerator
import serial
import re
import time


class QOL_LMX2487(RFGenerator):
    """
    QOL lab made RF generator based on TI LMX2487 chip.

    The communication with RF generator is done via nodeMCU controlling esp2866 chip.
    """

    def __init__(
        self,
        *args,
        port="/dev/ttyUSB0",
        speed=115200,
        timeout=1,
        setToDefaults=False,
        **kwds,
    ):
        super().__init__(*args, **kwds)
        self.config["Device model"] = "QOL made RF generator based on TI LMX2487 chip"
        self.port = port
        self.speed = speed
        self.timeout = timeout
        self.connection = serial.Serial(self.port, self.speed, timeout=self.timeout)
        self.log = []
        self.logCapacity = 10
        self._FreqFixed = None
        self.hopeFree = True
        self.hopeFreeFreqJump = 100e3  # we break laser lock if Freq change is larger
        self.dwellTime = 0.1  # needed for hope free setling
        if setToDefaults:
            self.sendSerialCmd("set_lmx2487_board_to_default_state()")

    def add2log(self, text):
        self.log.append(text)
        while len(self.log) > self.logCapacity:
            self.log.pop(0)

    def log2str(self, interval=None):
        strOut = ""
        for e in self.log:
            strOut += e
        return strOut

    def sendSerialCmd(self, cmd):
        self.connection.write(bytes(cmd + "\r", "ascii"))
        if "3.4" == serial.__version__:
            # older version style
            resp = self.connection.read_until(terminator=b"> ")
        else:
            # new style after 20180616
            resp = self.connection.read_until(expected=b"> ")
        resp = resp.decode("utf-8")
        self.add2log(resp)
        return resp

    @BasicInstrument.tsdb_append
    def setFreqFixed(self, freq):
        """Set frequency of RF signal.

        Will do incremental hope free tuning to desired frequency,
        if self.hopeFree is True.
        RF generator itself is fine, but our laser looses lock without it.
        """

        finished = False
        while not finished:
            frNow = self.getFreqFixed()
            dF = freq - frNow
            if dF >= 0:
                dFSign = 1
            else:
                dFSign = -1
            dF = abs(dF)
            if dF <= self.hopeFreeFreqJump or not self.hopeFree:
                fr = freq
                finished = True
            else:
                fr = frNow + dFSign * self.hopeFreeFreqJump
            self._FreqFixed = fr
            cmd_str = f"setFreq({fr:.2f})"
            self.sendSerialCmd(cmd_str)
            if not finished:
                time.sleep(self.dwellTime)

    @BasicInstrument.tsdb_append
    def getFreqFixed(self):
        """Gets RF signal frequency.

        Talking to hardware is slow, so we use cached value if we can.
        """
        if self._FreqFixed is None:
            resp = self.sendSerialCmd("getFreq()")
            m = re.search("[0-9.]+", resp)
            if m is not None:
                self._FreqFixed = float(m.group())
        return self._FreqFixed


if __name__ == "__main__":
    import platform

    if platform.system() == "Linux":
        rfgen = QOL_LMX2487(port="/dev/ttyUSB0", speed=115200, timeout=1)
    else:
        rfgen = QOL_LMX2487(port="COM4", speed=115200, timeout=1)
    print("testing")
    print("------ Header start -------------")
    print(str.join("\n", rfgen.getHeader()))
    print("------ Header ends  -------------")



"""
Module: HP 3457A Digital Multimeter Control
Description: This module provides a class to interface with a digital multimeter (DMM)
             using pyvisa for performing measurements such as voltage, current,
             and resistance.

Author: Mykhailo Vorobiov
Email: mvorobiov@wm.edu
Date: 2024-10-02
Updated: 2024-11-05
"""

from qolab.hardware.basic import BasicInstrument
from ._basic import Multimeter

import pyvisa
import numpy as np


class HP3457A(Multimeter):
    """
    A class to interface with HP 3457A Digital Multimeter (DMM) via pyvisa.

    This class allows you to measure DC voltage, AC voltage, DC current, AC current,
    and resistance from a DMM using pyvisa communication.
    """

    # List of data return formats
    FORMATS = ["ASCII", "SINT", "DINT", "SREAL", "DREAL"]

    BEEPER_STATUS = ["ON", "OFF", "ONCE"]

    # List of functions
    FUNCTIONS = [
        "DCV",
        "ACV",
        "ACDCV",  # (This is here for future work)
        "OHM",  # (This is here for future work)
        "OHMF",  # (This is here for future work)
        "DCI",
        "ACI",
        "ACDCI",  # (This is here for future work)
        "FREQ",  # (This is here for future work)
        "PER",  # (This is here for future work)
    ]

    # Allowed number of power line cycles (defines A/D integration time)
    NPLC = [0.0005, 0.005, 0.1, 1, 10, 100]

    # DC Volts Spec (This is here for future work)
    DCV_RES = {
        "30mv": {"6.5": 10e-9, "5.5": 100e-9, "4.5": 1e-6, "3.5": 10e-6},
        "300mv": {"6.5": 100e-9, "5.5": 1e-6, "4.5": 10e-6, "3.5": 100e-6},
        "3v": {"6.5": 1e-6, "5.5": 10e-6, "4.5": 100e-6, "3.5": 1e-3},
        "30v": {"6.5": 10e-6, "5.5": 100e-6, "4.5": 1e-3, "3.5": 10e-3},
        "300v": {"6.5": 100e-6, "5.5": 1e-3, "4.5": 10e-3, "3.5": 100e-3},
    }

    DCV_ACC = {
        "30mv": {
            "100": {"acc": 0.0045, "counts": 365},
            "10": {"acc": 0.0045, "counts": 385},
            "1": {"acc": 0.0045, "counts": 500},
            ".1": {"acc": 0.0045, "counts": 70},
            ".005": {"acc": 0.0045, "counts": 19},
            ".0005": {"acc": 0.0045, "counts": 6},
        },
        "300mv": {
            "100": {"acc": 0.0035, "counts": 39},
            "10": {"acc": 0.0035, "counts": 40},
            "1": {"acc": 0.0035, "counts": 50},
            ".1": {"acc": 0.0035, "counts": 9},
            ".005": {"acc": 0.0035, "counts": 4},
            ".0005": {"acc": 0.0035, "counts": 4},
        },
        "3v": {
            "100": {"acc": 0.0025, "counts": 6},
            "10": {"acc": 0.0025, "counts": 7},
            "1": {"acc": 0.0025, "counts": 7},
            ".1": {"acc": 0.0025, "counts": 4},
            ".005": {"acc": 0.0025, "counts": 4},
            ".0005": {"acc": 0.0025, "counts": 4},
        },
        "30v": {
            "100": {"acc": 0.0040, "counts": 19},
            "10": {"acc": 0.0040, "counts": 20},
            "1": {"acc": 0.0040, "counts": 30},
            ".1": {"acc": 0.0040, "counts": 7},
            ".005": {"acc": 0.0040, "counts": 4},
            ".0005": {"acc": 0.0040, "counts": 4},
        },
        "300v": {
            "100": {"acc": 0.0055, "counts": 6},
            "10": {"acc": 0.0055, "counts": 7},
            "1": {"acc": 0.0055, "counts": 7},
            ".1": {"acc": 0.0055, "counts": 4},
            ".005": {"acc": 0.0055, "counts": 4},
            ".0005": {"acc": 0.0055, "counts": 4},
        },
    }

    # DC Current Spec (This is here for future work)
    DCI_RES = {
        "300ua": {"6.5": 100e-12, "5.5": 1e-9, "4.5": 10e-9, "3.5": 100e-9},
        "3ma": {"6.5": 1e-9, "5.5": 10e-9, "4.5": 100e-9, "3.5": 1e-6},
        "30ma": {"6.5": 10e-9, "5.5": 100e-9, "4.5": 1e-6, "3.5": 10e-6},
        "300ma": {"6.5": 100e-9, "5.5": 1e-6, "4.5": 10e-6, "3.5": 100e-6},
        "1a": {"6.5": 1e-6, "5.5": 10e-6, "4.5": 100e-6, "3.5": 1e-3},
    }
    # (This is here for future work)
    DCI_ACC = {
        "300ua": {
            "100": {"acc": 0.04, "counts": 104},
            "10": {"acc": 0.04, "counts": 104},
            "1": {"acc": 0.04, "counts": 115},
            ".1": {"acc": 0.04, "counts": 14},
            ".005": {"acc": 0.04, "counts": 5},
            ".0005": {"acc": 0.04, "counts": 4},
        },
        "3ma": {
            "100": {"acc": 0.04, "counts": 104},
            "10": {"acc": 0.04, "counts": 104},
            "1": {"acc": 0.04, "counts": 115},
            ".1": {"acc": 0.04, "counts": 14},
            ".005": {"acc": 0.04, "counts": 5},
            ".0005": {"acc": 0.04, "counts": 4},
        },
        "30ma": {
            "100": {"acc": 0.04, "counts": 104},
            "10": {"acc": 0.04, "counts": 104},
            "1": {"acc": 0.04, "counts": 115},
            ".1": {"acc": 0.04, "counts": 14},
            ".005": {"acc": 0.04, "counts": 5},
            ".0005": {"acc": 0.04, "counts": 4},
        },
        "300ma": {
            "100": {"acc": 0.08, "counts": 204},
            "10": {"acc": 0.08, "counts": 204},
            "1": {"acc": 0.08, "counts": 215},
            ".1": {"acc": 0.08, "counts": 24},
            ".005": {"acc": 0.08, "counts": 6},
            ".0005": {"acc": 0.08, "counts": 4},
        },
        "1a": {
            "100": {"acc": 0.08, "counts": 604},
            "10": {"acc": 0.08, "counts": 604},
            "1": {"acc": 0.08, "counts": 615},
            ".1": {"acc": 0.08, "counts": 64},
            ".005": {"acc": 0.08, "counts": 10},
            ".0005": {"acc": 0.08, "counts": 5},
        },
    }

    def __init__(self, resource, *args, **kwds):
        """
        Initialize the Digital Multimeter class and establish a connection.

        :param resource_name: VISA resource name for the multimeter.
        :param alias: Alias for logging.
        :param log_level: Logging level (default: INFO).
        """
        super().__init__(*args, **kwds)
        self.resource = resource
        self.config["Device model"] = "HP 3457A"
        self.resource.timeout = 5000
        self.resource.read_termination = "\r\n"
        self.resource.write_termination = "\r\n"

        self.read = self.resource.read
        self.write = self.resource.write
        self.read_bytes = self.resource.read_bytes

        self.switchTime = 0.5  # switch time in seconds for Function/Measurement change
        self.deviceProperties.update({"Function"})
        self.set_function("DCV")

    def get_idn(self):
        """
        Query the identification of the digital multimeter: manufacturer, model, serial number, and firmware version.

        :return: Identifier string if query is successful, or None otherwise.
        """
        try:
            self.write("ID?")
            return self.read_bytes(9).decode().strip()
        except pyvisa.VisaIOError as e:
            print(
                f"[ERROR!] Error retrieving identification of the digital multimeter: {e}"
            )
            return None

    def _check_format(self, format):
        if format in self.FORMATS:
            return True
        else:
            return False

    def set_format(self, format="ASCII"):
        if self._check_format(format):
            try:
                self.write(f"OFORMAT {format}")
                print(f"[INFO] Format  has been set to {format}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Couldn't set reading format to {format}: {e}")
        else:
            print(
                f"[WARNING] Entered format is not allowed. Allowed formats: {self.FORMATS}"
            )

    @BasicInstrument.tsdb_append
    def get_reading(self):
        try:
            self.write("TARM AUTO")
            return float(self.read_bytes(16).decode().strip())
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error failed to get reading: {e}")
            return np.nan()

    def set_beeper_status(self, status="OFF"):
        if status in self.BEEPER_STATUS:
            try:
                self.write(f"BEEP {status}")
                print(f"[INFO] Beeper is set to {status}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting beeper status: {e}")
        else:
            print(
                f"[WARNING!] The passed beeper status is not allowed."
                f"Choose from {self.BEEPER_STATUS}"
            )

    def set_function(self, function="DCV"):
        """
        Set the measurement function of the multimeter.

        :param function: Measurement function .
        """
        if function in self.FUNCTIONS:
            try:
                self.write(f"FUNC {function}")
                print(f"[INFO] Measurement function set to {function}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting measurement function: {e}")
        else:
            print(
                f"[WARNING!] Invalid measurement function '{function}'. "
                f"Valid functions are: {', '.join(self.FUNCTIONS)}"
            )

    @BasicInstrument.tsdb_append
    def get_temperature(self):
        try:
            self.write("TEMP?")
            temp = float(self.read(16).decode().strip())
            print(f"[INFO] DMM's internal temperature is: {temp}")
            return temp
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error retrieving DMM's internal temperature: {e}")

    def get_memory_count(self):
        """
        Returns the total number of stored readings.
        """
        try:
            self.write("MCOUNT?")
            return float(self.read_bytes(16).decode().strip())
        except pyvisa.VisaIOError as e:
            print(f"[ERROR!] Error failed to get count of reading in memory: {e}")
            return np.nan()

    def enable_keyboard(self, status=True):
        if status:
            try:
                self.write("LOCK ON")
                print("[INFO] DMM's keyboard is unlocked.")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error unlocking the DMM's keyboard: {e}")
        else:
            try:
                self.write("LOCK OFF")
                print("[INFO] DMM's keyboard is locked.")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error locking the DMM's keyboard: {e}")

    def set_nplc(self, nplc=10):
        if nplc in self.NPLC:
            try:
                self.write(f"FUNC {nplc}")
                print(f"[INFO] NPLC is set to {nplc}")
            except pyvisa.VisaIOError as e:
                print(f"[ERROR!] Error setting number of power line cycles (NPLC): {e}")
        else:
            print(
                f"[WARNING!] Invalid NPLC '{nplc}'. "
                f"Valid NPLC values are: {', '.join(str(self.NPLC))}"
            )

    # -----------------------------------------------------------
    # Methods to comply with other DMM QOLab module drivers
    # -----------------------------------------------------------
    @BasicInstrument.tsdb_append
    def getVdc(self):
        self.set_function("DCV")
        return self.get_reading()

    @BasicInstrument.tsdb_append
    def getVac(self):
        self.set_function("ACV")
        return self.get_reading()

    @BasicInstrument.tsdb_append
    def getAdc(self):
        self.set_function("DCI")
        return self.get_reading()

    @BasicInstrument.tsdb_append
    def getAac(self):
        self.set_function("ACI")
        return self.get_reading()


# Example usage
if __name__ == "__main__":

    rm = pyvisa.ResourceManager()
    instrument = rm.open_resource("visa://192.168.194.15/GPIB1::22::INSTR")
    dmm = HP3457A(instrument)
    dmm.set_format("ASCII")
    dmm.set_beeper_status("ONCE")
    dmm.set_function("DCI")
    print(dmm.get_reading())
    dmm.close()
